#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod

from .model_index import ModelIndex

from ..core.signaling import Signal


#------------------------------------------------------------------------------
# The various AbstractItemModel flags
#------------------------------------------------------------------------------
# Item flags (these are the same values as Qt's flags so we don't
# have to do a map lookup for Qt).
NO_ITEM_FLAGS = 0x0         # The item has no state flags
ITEM_IS_SELECTABLE = 0x1    # The item can be selected in a view.
ITEM_IS_EDITABLE = 0x2      # The item can be edited in a view.
ITEM_IS_DRAG_ENABLED = 0x4  # The item can be dragged.
ITEM_IS_DROP_ENABLED = 0x8  # Another item can be dropped onto this one.
ITEM_IS_CHECKABLE = 0x10    # The item's "checked" state can be changed from a view.
ITEM_IS_ENABLED = 0x20      # This item permits user interaction in a view.
ITEM_IS_TRISTATE = 0x40     # The item is checkable, and it has three distinct states.


# Alignment flags (these are the same values as Qt's flags so we don't
# have to do a map lookup for Qt).
ALIGN_LEFT = 0x1
ALIGN_RIGHT = 0x2
ALIGN_HCENTER = 0x4
ALIGN_JUSTIFY = 0x8
ALIGN_TOP = 0x20
ALIGN_BOTTOM = 0x40
ALIGN_VCENTER = 0x80
ALIGN_CENTER = ALIGN_HCENTER | ALIGN_VCENTER


#------------------------------------------------------------------------------
# AbstractItemModel
#------------------------------------------------------------------------------
class AbstractItemModel(object):
    """ An abstract model for supplying information to heierarchical
    widgets.

    """
    __metaclass__ = ABCMeta

    #: Fired by the begin_insert_columns method
    columns_about_to_be_inserted = Signal()

    #: Fired by the begin_move_columns method
    columns_about_to_be_moved = Signal()

    #: Fired by the begin_remove_columns method
    columns_about_to_be_removed = Signal()

    #: Fired by the end_insert_columns method
    columns_inserted = Signal()

    #: Fired by the end_move_columns method
    columns_moved = Signal()

    #: Fired by the end_remove_rows method
    columns_removed = Signal()

    #: Fired by the begin_insert_rows method
    rows_about_to_be_inserted = Signal()

    #: Fired by the begin_move_rows method
    rows_about_to_be_moved = Signal()

    #: Fired by the begin_remove_columns method
    rows_about_to_be_removed = Signal()

    #: Fired by the end_insert_rows method
    rows_inserted = Signal()

    #: Fired by the end_move_rows method
    rows_moved = Signal()

    #: Fired by the end_remove_rows method
    rows_removed = Signal()

    #: Fired by the begin_change_layout method
    layout_about_to_be_changed = Signal()

    #: Fired by the end_change_layout method
    layout_changed = Signal()

    #: Fired by the begin_reset_model method
    model_about_to_be_reset = Signal()

    #: Fired by the end_reset_model method
    model_reset = Signal()

    #: Fired by the notify_data_changed method
    data_changed = Signal()
    
    #: Fired by the notify_horizontal_header_data_changed method
    horizontal_header_data_changed = Signal()

    #: Fired by the notify_vertical_header_data_changed method
    vertical_header_data_changed = Signal()
    
    #--------------------------------------------------------------------------
    # Model change notification trigger methods 
    #--------------------------------------------------------------------------
    def begin_insert_columns(self, parent, first, last):
        """ Begins inserting a column.

        This method must be called before inserting data into
        the model.

        Arguments
        ---------
        parent : ModelIndex
            The parent into which the new columns will be inserted.

        first : int
            The column position at which insertion will begin.

        last : int
            The column position at which insertion will end.

        """
        evt_arg = (parent, first, last)
        self.columns_about_to_be_inserted(evt_arg)

    def begin_move_columns(self, src_parent, src_first, src_last, dst_parent, dst_child):
        """ Begins to move a column.

        Arguments
        ---------
        src_parent : ModelIndex
            The item from which columns will be moved.

        src_first : int
            The number of the first column to be moved.

        src_last : int
            The number of the last column to be moved.

        dst_parent : ModelIndex
            The item into which the columns will be moved.

        dst_child : int
            The column number to which the columns will be moved.

        """
        evt_arg = (src_parent, src_first, src_last, dst_parent, dst_child)
        self.columns_about_to_be_moved(evt_arg)

    def begin_remove_columns(self, parent, first, last):
        """ Begins to remove columns.

        This method must be called before removing data from the
        model.

        Arguments
        ---------
        parent : ModelIndex
            The item from which the columns will be removed.

        first : int
            The number of the first column to remove.

        last : int
            The number of the last column to remove.

        """
        evt_arg = (parent, first, last)
        self.columns_about_to_be_removed(evt_arg)

    def end_insert_columns(self, parent, first, last):
        """ Finish inserting columns.

        This method must be called after inserting data into
        the model.

        Arguments
        ---------
        parent : ModelIndex
            The parent into which the new columns will be inserted.

        first : int
            The column position at which insertion will begin.

        last : int
            The column position at which insertion will end.

        """
        evt_arg = (parent, first, last)
        self.columns_inserted(evt_arg)

    def end_move_columns(self, src_parent, src_first, src_last, dst_parent, dst_child):
        """ Finish moving columns.

        This method must be called after moving data in a model.

        Arguments
        ---------
        src_parent : ModelIndex
            The item from which columns will be moved.

        src_first : int
            The number of the first column to be moved.

        src_last : int
            The number of the last column to be moved.

        dst_parent : ModelIndex
            The item into which the columns will be moved.

        dst_child : int
            The column number to which the columns will be moved.

        """
        evt_arg = (src_parent, src_first, src_last, dst_parent, dst_child)
        self.columns_moved(evt_arg)

    def end_remove_columns(self, parent, first, last):
        """ Finish removing columns.

        This method must be called after removing data from a model.

        Arguments
        ---------
        parent : ModelIndex
            The item from which the columns will be removed.

        first : int
            The number of the first column to remove.

        last : int
            The number of the last column to remove.

        """
        evt_arg = (parent, first, last)
        self.columns_removed(evt_arg)

    def begin_insert_rows(self, parent, first, last):
        """ Begins a row insertion.

        This method must be called before inserting data into
        the model.

        Arguments
        ---------
        parent : ModelIndex
            The item into which the new rows will be inserted.

        first : int
            The row position at which insertion will begin.

        last : int
            The row position at which insertion will end.

        """
        evt_arg = (parent, first, last)
        self.rows_about_to_be_inserted(evt_arg)

    def begin_move_rows(self, src_parent, src_first, src_last,
                        dst_parent, dst_child):
        """ Begins to move a row.

        Arguments
        ---------
        src_parent : ModelIndex
            The item from which rows will be moved.

        src_first : int
            The number of the first row to be moved.

        src_last : int
            The number of the last row to be moved.

        dst_parent : ModelIndex
            The item into which the rows will be moved.

        dst_child : int
            The row number to which the rows will be moved.

        """
        evt_arg = (src_parent, src_first, src_last, dst_parent, dst_child)
        self.rows_about_to_be_moved(evt_arg)

    def begin_remove_rows(self, parent, first, last):
        """ Begins to remove rows.

        This method must be called before removing data from the
        model.

        Arguments
        ---------
        parent : ModelIndex
            The item from which the rows will be removed.

        first : int
            The number of the first row to remove.

        last : int
            The number of the last row to remove.

        """
        evt_arg = (parent, first, last)
        self.rows_about_to_be_removed(evt_arg)

    def end_insert_rows(self, parent, first, last):
        """ Finish inserting rows.

        This method must be called after inserting data into
        the model.

        Arguments
        ---------
        parent : ModelIndex
            The item into which the new rows will be inserted.

        first : int
            The row position at which insertion will begin.

        last : int
            The row position at which insertion will end.

        """
        evt_arg = (parent, first, last)
        self.rows_inserted(evt_arg)

    def end_move_rows(self, src_parent, src_first, src_last, dst_parent, dst_child):
        """ Finish moving rows

        This method must be called after moving data in a model.

        Arguments
        ---------
        src_parent : ModelIndex
            The item from which rows will be moved.

        src_first : int
            The number of the first row to be moved.

        src_last : int
            The number of the last row to be moved.

        dst_parent : ModelIndex
            The item into which the rows will be moved.

        dst_child : int
            The row number to which the rows will be moved.

        """
        evt_arg = (src_parent, src_first, src_last, dst_parent, dst_child)
        self.rows_moved(evt_arg)

    def end_remove_rows(self, parent, first, last):
        """ Finish removing rows.

        This method must be called after moving data in a model.

        Arguments
        ---------
        parent : ModelIndex
            The item from which the rows will be removed.

        first : int
            The number of the first row to remove.

        last : int
            The number of the last row to remove.

        """
        evt_arg = (parent, first, last)
        self.rows_removed(evt_arg)

    def begin_change_layout(self):
        """ Begin a change to the layout of the model -- e.g., sort it.

        This method must be called before rearranging data in a model.

        """
        self.layout_about_to_be_changed()

    def end_change_layout(self):
        """ Finish a change to the layout of the model.

        This method must be called after rearranging data in a model.

        """
        self.layout_changed()

    def begin_reset_model(self):
        """ Begin to reset a model.

        Model indices must be recomputed, and any associated views will 
        also be reset.

        This method must be called before a model is reset.

        """
        self.model_about_to_be_reset()

    def end_reset_model(self):
        """ Finish resetting a model.

        This method must be called after a model is reset.

        """
        self.model_reset()

    def notify_data_changed(self, top_left, bottom_right):
        """ Create a notification event for the model data has changed.

        Arguments
        ---------
        top_left : ModelIndex
            The upper-left boundary of the changed items.

        bottom_right : ModelIndex
            The bottom-right boundary of the changed items.

        """
        self.data_changed((top_left, bottom_right))

    def notify_horizontal_header_data_changed(self, first, last):
        """ Create a notification that model horizontal header data 
        has changed.

        Arguments
        ---------
        first : int
            The first horizontal/column header that has been modified.

        last : int
            The last horizontal/column header that has been modified.

        """
        self.horizontal_header_data_changed((first, last))

    def notify_vertical_header_data_changed(self, first, last):
        """ Create a notification that model vertical header data 
        has changed.

        Arguments
        ---------
        first : int
            The first vertical/row header that has been modified.

        last : int
            The last vertical/row header that has been modified.

        """
        self.vertical_header_data_changed((first, last))

    #--------------------------------------------------------------------------
    # Misc methods 
    #--------------------------------------------------------------------------
    def sibling(self, row, column, index):
        """ Obtain an item with the same parent as `index`.

        Arguments
        ---------
        row : int
            The row of the new item.

        column : int
            The column of the new item.

        index : ModelIndex
            An item for which to find a "sibling".

        Returns
        -------
        index : ModelIndex
            An element with the same parent as the specified item.

        """
        return self.index(row, column, self.parent(index))

    def buddy(self, index):
        """ Refers the caller to an item in the model to edit.

        Each item is its own buddy by default, but this is not required.

        Arguments
        ---------
        index : ModelIndex
            The model index for which the buddy index should be 
            returned for editing.

        Returns
        -------
        buddy : ModelIndex
            The item that should be edited.

        """
        return index

    def can_fetch_more(self, parent=None):
        """ Returns `True` if items of `parent` can provide more data.

        Arguments
        ---------
        parent : ModelIndex or None
            A model index, possibly with more data that can be fetched.

        Returns
        -------
        has_more : bool
            True if more data is available, False otherwise.

        """
        return False

    def fetch_more(self, parent=None):
        """ Obtain data from model items, if more is available.

        This method is useful when incrementally adding data to a model.

        Arguments
        ---------
        parent : ModelIndex or None
            An index to query.

        Returns
        -------
        data : AbstractItemModel
            Data that has yet to be received.

        """
        pass

    def has_index(self, row, column, parent=None):
        """ Determine whether this model can provide a valid index with
        certain specifications.

        Arguments
        ---------
        row : int
            A row to be used in the index.

        column : int
            A column to be used in the index.

        parent : ModelIndex or None
            A parent on which to base the new model index.

        Returns
        -------
        result : bool
            True if the inputs produce a valid index, False otherwise.

        """
        if row < 0 or column < 0:
            return False
        return row < self.row_count(parent) and column < self.column_count(parent)

    def has_children(self, parent=None):
        """ Determines if an index has any child items.

        Arguments
        ---------
        parent : ModelIndex or None
            A model index to check for children.

        Returns
        -------
        result : bool
            True if the index has children, False otherwise.

        """
        return self.row_count(parent) > 0 and self.column_count(parent) > 0

    def create_index(self, row, column, context):
        """ Create a new index into this model.

        Arguments
        ---------
        row : int
            A row to specify the index

        column : int
            A column to specify the index.

        context : object
            A value that can be useful when working with a model.

        Returns
        -------
        index : ModelIndex
            A new index into this model.

        """
        return ModelIndex(row, column, context, self)

    def flags(self, index):
        """ Obtain the flags that specify user interaction with items.

        Arguments
        ---------
        index : ModelIndex
            Items to query.

        Returns
        -------
        item_flags : tuple of strings
            The ways in which a view can interact with the items.

        """
        return ITEM_IS_SELECTABLE | ITEM_IS_ENABLED

    #--------------------------------------------------------------------------
    # Abstract Methods
    #--------------------------------------------------------------------------
    @abstractmethod
    def column_count(self, parent=None):
        """ Count the number of columns in the children of an item.

        Arguments
        ---------
        parent : ModelIndex or None
            A model item with children.

        Returns
        -------
        count : int
            The number of columns in the model.

        """
        raise NotImplementedError

    @abstractmethod
    def row_count(self, parent=None):
        """ Count the number of rows in the children of an item.

        Arguments
        ---------
        parent : ModelIndex or None
            A model item with children.

        Returns
        -------
        count : int
            The number of rows in the model.

        """
        raise NotImplementedError

    @abstractmethod
    def index(self, row, column, parent=None):
        """ Obtain an index coresponding to an item in the model.

        Arguments
        ---------
        row : int
            The sought-after item's row.

        column : int
            The sought-after item's column.

        parent : ModelIndex or None
            The parent item that is the base of this index.

        Returns
        -------
        item : ModelIndex or None
            An index for the specified item.

        """
        raise NotImplementedError

    @abstractmethod
    def parent(self, index):
        """ Obtain the parent of a model item.

        Arguments
        ---------
        index : ModelIndex
            A model item, possibly with a valid parent.

        Returns
        -------
        index : ModelIndex
            An index for the parent item.

        """
        raise NotImplementedError


    @abstractmethod
    def data(self, index):
        """ Get the data for a model index as a string for display.

        Arguments
        ---------
        index : ModelIndex
            The model index for which to return the data.

        Returns
        -------
        value : string
            The data as a string for display.

        """
        raise NotImplementedError

    #--------------------------------------------------------------------------
    # Auxiliary Data Methods
    #--------------------------------------------------------------------------
    def decoration(self, index):
        """ Return an icon for the index

        The default implementation returns None.

        Arguments
        ---------
        index : ModelIndex
            The model index for which to return the icon.

        Returns
        -------
        value : icon??
            The icon to display for the cell.

        """
        # XXX Handle Icons !!!
        return None
    
    def edit_data(self, index):
        """ Get the data for a model index as a string for editing.

        The default implementation returns 'data(index)'.

        Arguments
        ---------
        index : ModelIndex
            The model index for which to return the data.

        Returns
        -------
        value : string
            The data as a string for editing.

        """
        return self.data(index)
    
    def tool_tip(self, index):
        """ Get the tool tip string for a given model index.

        The default implementation returns None.

        Arguments
        ---------
        index : ModelIndex
            The model index for which to return the tool tip.

        Returns
        -------
        value : string
            The tool tip string.

        """
        return None
    
    def status_tip(self, index):
        """ Get the status tip string for a given model index.

         The default implementation returns None.

        Arguments
        ---------
        index : ModelIndex
            The model index for which to return the status tip.

        Returns
        -------
        value : string
            The status tip string.

        """
        return None

    def whats_this(self, index):
        """ Get the what's this? string for a given model index.

         The default implementation returns None.

        Arguments
        ---------
        index : ModelIndex
            The model index for which to return the what's this? string.

        Returns
        -------
        value : string
            The what's this? string.

        """
        return None
    
    def font(self, index):
        """ Get the font for a given model index.

        The default implementation returns None.

        Arguments
        ---------
        index : ModelIndex
            The model index for which to return the font.

        Returns
        -------
        value : Font
            The font to use for this model index.

        """
        return None
    
    def alignment(self, index):
        """ Get the alignment of the text for a given model index.

        The default implementation returns ALIGN_VCENTER | ALIGN_RIGHT

        Arguments
        ---------
        index : ModelIndex
            The model index for which to return the alignment.

        Returns
        -------
        value : int
            The OR'd alignment flags to use for the given model index.

        """
        return ALIGN_VCENTER | ALIGN_RIGHT
    
    def background(self, index):
        """ The background brush to use for the given index.

        The default implementation returns None.

        Arguments
        ---------
        index : ModelIndex
            The model index for which to return the brush.

        Returns
        -------
        value : Brush
            The brush to use for the given model index.

        """
        return None
    
    def foreground(self, index):
        """ The foreground brush to use for the given index.

        The default implementation returns None.

        Arguments
        ---------
        index : ModelIndex
            The model index for which to return the brush.

        Returns
        -------
        value : Brush
            The brush to use for the given model index.

        """
        return None
    
    def check_state(self, index):
        """ Get the check state for the given index.

        The default implementation returns None.

        Arguments
        ---------
        index : ModelIndex
            The model index for which to return the check state.

        Returns
        -------
        value : string
            The check state to use for the given model index.

        """
        return None
    
    def size_hint(self, index):
        """ Get the (width, height) size hint for the given index.

        The default implementation returns None.

        Arguments
        ---------
        index : ModelIndex
            The model index for which to return the size hint.

        Returns
        -------
        value : (width, height)
            The size hint to use for the given model index.

        """
        return None

    #--------------------------------------------------------------------------
    # Data Setter Methods
    #--------------------------------------------------------------------------
    def set_data(self, index, value):
        """ Update a model item's data. The default implementation does
        nothing and returns False.

        Implementations that allow editing must set the 'editable' flag
        to True, and call notify_data_changed(...) explicity in order
        for the table to update.

        Arguments
        ---------
        index : ModelIndex
            The model index for the item to update.

        value : object
            The user supplied value to set in the model.

        Returns
        -------
        success : bool
            True if the data was successfully set, False otherwise.

        """
        return False
    
    def set_check_state(self, index, value):
        """ Update a model item's check state. The default implementation 
        does nothing and returns False.

        Implementations that allow editing must set the 'editable' flag
        to True, and call notify_data_changed(...) explicity in order
        for the table to update.

        Arguments
        ---------
        index : ModelIndex
            The model index for the item to update.

        value : object
            The check state to set in the model.

        Returns
        -------
        success : bool
            True if the data was successfully set, False otherwise.

        """
        return False
    
    #--------------------------------------------------------------------------
    # Header Data Methods 
    #--------------------------------------------------------------------------
    def horizontal_header_data(self, section):
        """ Get the horizontal label for a particular header section.

        The default implementation returns the column number of the given 
        header. Subclasses should override this method to return 
        appropriate data for their model.

        Arguments
        ---------
        section : int
            The column number of the header.

        Returns
        -------
        data : string
            The requested header label

        """
        return unicode(section + 1)
    
    def vertical_header_data(self, section):
        """ Get the vertical label for a particular header section.

        The default implementation returns the row number of the given 
        header. Subclasses should override this method to return 
        appropriate data for their model.

        Arguments
        ---------
        section : int
            The row number of the header.

        Returns
        -------
        data : string
            The requested header label

        """
        return unicode(section + 1)

    #--------------------------------------------------------------------------
    # Header Data auxiliary methods
    #--------------------------------------------------------------------------
    def horizontal_header_decoration(self, section):
        """ Get the horizontal icon for a particular header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The column number of the header.

        Returns
        -------
        data : icon??
            The requested icon for the header

        """
        # XXX implement icons
        return None
    
    def vertical_header_decoration(self, section):
        """ Get the vertical icon for a particular header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The row number of the header.

        Returns
        -------
        data : icon??
            The requested icon for the header

        """
        # XXX implement icons
        return None

    def horizontal_header_tool_tip(self, section):
        """ Get the horizontal tool tip for a particular header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The column number of the header.

        Returns
        -------
        data : string
            The requested tool tip for the header

        """
        return None
    
    def vertical_header_tool_tip(self, section):
        """ Get the vertical tool tip for a particular header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The row number of the header.

        Returns
        -------
        data : string
            The requested tool tip for the header

        """
        return None   
    
    def horizontal_header_status_tip(self, section):
        """ Get the horizontal status tip for a particular header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The column number of the header.

        Returns
        -------
        data : string
            The requested status tip for the header

        """
        return None
    
    def vertical_header_status_tip(self, section):
        """ Get the vertical status tip for a particular header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The row number of the header.

        Returns
        -------
        data : string
            The requested status tip for the header

        """
        return None 
        
    def horizontal_header_whats_this(self, section):
        """ Get the horizontal whats this for a particular header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The column number of the header.

        Returns
        -------
        data : string
            The requested what's this? for the header

        """
        return None
    
    def vertical_header_whats_this(self, section):
        """ Get the vertical whats this for a particular header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The row number of the header.

        Returns
        -------
        data : string
            The requested what's this? for the header

        """
        return None 
    
    def horizontal_header_font(self, section):
        """ Get the horizontal font for a particular header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The column number of the header.

        Returns
        -------
        data : Font()
            The requested font for the header

        """
        return None
    
    def vertical_header_font(self, section):
        """ Get the vertical font for a particular header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The row number of the header.

        Returns
        -------
        data : Font()
            The requested font for the header

        """
        return None 

    def horizontal_header_alignment(self, section):
        """ Get the horizontal whats this for a particular header section.

        The default implementation returns ALIGN_CENTER.

        Arguments
        ---------
        section : int
            The column number of the header.

        Returns
        -------
        data : int
            The OR'd alignment flags for the header

        """
        return ALIGN_CENTER
    
    def vertical_header_alignment(self, section):
        """ Get the vertical whats this for a particular header section.

        The default implementation returns ALIGN_CENTER

        Arguments
        ---------
        section : int
            The row number of the header.

        Returns
        -------
        data : int
            The OR'd alignment flags for the header

        """
        return ALIGN_CENTER
    
    def horizontal_header_background(self, section):
        """ Get the horizontal background brush for a particular 
        header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The column number of the header.

        Returns
        -------
        data : Brush()
            The requested brush for the header

        """
        return None
    
    def vertical_header_background(self, section):
        """ Get the vertical background brush for a particular
        header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The row number of the header.

        Returns
        -------
        data : Brush()
            The requested brush for the header

        """
        return None 

    def horizontal_header_foreground(self, section):
        """ Get the horizontal foreground brush for a particular 
        header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The column number of the header.

        Returns
        -------
        data : Brush()
            The requested brush for the header

        """
        return None
    
    def vertical_header_foreground(self, section):
        """ Get the vertical foreground brush for a particular
        header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The row number of the header.

        Returns
        -------
        data : Brush()
            The requested brush for the header

        """
        return None

    def horizontal_header_size_hint(self, section):
        """ Get the size hint for a particular horizontal header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The column number of the header.

        Returns
        -------
        value : (width, height)
            The size hint to use for the given column number.

        """
        return None
    
    def vertical_header_size_hint(self, section):
        """ Get the size hint for a particular vertical header section.

        The default implementation returns None.

        Arguments
        ---------
        section : int
            The row number of the header.

        Returns
        -------
        value : (width, height)
            The size hint to use for the given row number.

        """
        return None

    #--------------------------------------------------------------------------
    # Header Data Setters
    #--------------------------------------------------------------------------
    
    # At the moment we dont have a compelling use case where we
    # need to be able to set the header data from the ui. But in 
    # the future, that functionality will be added here.


#------------------------------------------------------------------------------
# AbstractTableModel
#------------------------------------------------------------------------------
class AbstractTableModel(AbstractItemModel):
    """ An AbstractItemModel subclass that implements the methods 
    necessary to restrict the model to a 2-dimensional table of data.

    """
    def index(self, row, column, parent=None):
        """ Abstract method implementation that will create a valid
        model index for the given row and column if they are valid.
        Otherwise, returns an invalid index.

        """
        if self.has_index(row, column, parent):
            return self.create_index(row, column, None)

    def parent(self, index):
        """ Abstract method implementation that always returns None, 
        forcing the model to be flat.

        """
        return None

    def has_children(self, parent):
        """ Overridden parent class method which restrics the model to
        2-dimensions.

        """
        if parent is None or parent.model is self:
            return self.row_count(parent) > 0 and self.column_count(parent) > 0
        return False


#------------------------------------------------------------------------------
# AbstractListModel
#------------------------------------------------------------------------------
class AbstractListModel(AbstractItemModel):
    """ An AbstractItemModel subclass that implements the methods 
    necessary to restrict the model to a 1-dimensional list of data.

    """
    def index(self, row, column, parent=None):
        """ Abstract method implementation that will create a valid
        model index for the given row and column if they are valid.
        Otherwise, returns an invalid index.

        """
        if self.has_index(row, column, parent):
            return self.create_index(row, column, None)

    def parent(self, index):
        """ Abstract method implementation that always returns an
        invalid index, forcing the model to be flat.

        """
        return None

    def column_count(self, parent=None):
        """ Overridden parent class method which restricts the model
        to a single column.

        """
        if parent is not None:
            return 0
        return 1

    def has_children(self, parent=None):
        """ Overridden parent class method which restrics the model
        to 1-dimension.

        """
        if parent is not None:
            return False
        return self.row_count() > 0




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import math
import random
import datetime
Window:

    title << model.window_title

    Panel:
        HGroup:

            style.cls = "stretch"

            Group my_group:
                direction = 'top_to_bottom'
                PushButton:
                    style.cls = "expanding"
                    text << "clickme!" if not self.down else "I'm down!"
                    clicked >> model.update_style()

                PushButton pb2:
                    style.cls = "expanding"
                    text = "shuffle"
                    clicked >> setattr(rb1, 'checked', True)

                PushButton static:
                    style.cls = "expanding error_colors"
                    text = "static"
                    clicked >> model.print_msg(msg)

                CheckBox cb1:
                    text = "A simple text box"
                    toggled >> setattr(self, 'text', model.randomize(self.text))


            Html:
                source << ("<center><h1>Hello Enaml!</h1></center><br>" * cmbx.value
                           if not static.down else
                           "<center><h1>Static Down!</h1></center>")

            VGroup:
                style.cls = "fixed"
                Panel:
                    HGroup:
                        direction = 'right_to_left'
                        RadioButton rb1:
                            style.cls = "stretch"
                            text = 'rb1'
                            toggled >> print('rb1:', self.checked)
                        RadioButton:
                            style.cls = "stretch"
                            text = 'rb2'
                            toggled >> print('rb2:', self.checked)
                        RadioButton:
                            style.cls = "stretch"
                            text = 'rb3'
                            toggled >> print('rb3:', self.checked)
                        RadioButton:
                            style.cls = "stretch"
                            text = 'rb4'
                            toggled >> print('rb4:', self.checked)
                Panel:
                    HGroup:
                        RadioButton:
                            style.cls = "stretch"
                            text = 'rb1'
                        RadioButton:
                            style.cls = "stretch"
                            text = 'rb2'
                        RadioButton:
                            style.cls = "stretch"
                            text = 'rb3'
                        RadioButton:
                            style.cls = "stretch"
                            text = 'rb4'

                Calendar:
                    minimum_date = datetime.date(1970, 1, 1)
                    activated >> print('activated', msg.new)
                    selected >> print('selected', msg.new)
                    date >> print('new date', msg.new)

                HGroup:
                    ComboBox cmbx:
                        items = range(100)
                        value = 1
                        selected >> print('selected', msg.new)

                HGroup:
                    style.border_left_width = 20
                    style.border_top_width = 20
                    style.border_bottom_width = 20
                    Label:
                        text = 'min:'
                    Label:
                        text << str(sb.low)
                    Label:
                        text = 'max:'
                    Label:
                        text << str(sb.high)
                    Label:
                        text = 'val:'
                    Label:
                        text << str(sb.value)

                HGroup:
                    CheckBox wrap_box:
                        style.cls << "error_colors" if mf.error else "normal_colors"
                        text = 'Allow wrap:'
                        toggled >> setattr(cmbx, 'value', 42)
                    SpinBox sb:
                        prefix << 'Foo ' if wrap_box.checked else 'Bar '
                        suffix << ' lb' if wrap_box.checked else ' kg'
                        wrap := wrap_box.checked
                        special_value_text = "Auto"
                        step = 2
                        low << -20 if not self.wrap else 0
                        high = 20

                HGroup:
                    Label:
                        text = "A Slider!"
                        style.cls = "no_stretch"
                    Slider:
                        value = 10
                        value >> print(self.value)
                        to_slider = lambda val: math.log10(val)
                        from_slider = lambda val: 10 ** val
                        tick_interval = 0.5
                        orientation << ('vertical'
                                        if wrap_box.checked else
                                        'horizontal')

                Field mf:
                    to_string = str
                    from_string = lambda val: int(val) if val else 0
                    value = 0
                    style.background_color << "error" if self.error else "nocolor"



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import datetime

from traits.api import HasTraits, Str, Int, Instance, Tuple, Date, Property


class Person(HasTraits):
    """ A simple class representing a person object.

    """
    # The last name of the person as a string
    last_name = Str

    # The first name of the person as a string
    first_name = Str

    # The date of birth of the person
    dob = Date(datetime.date(1970, 1, 1))

    # The age of the person computed from their dob
    age = Property(Int, depends_on='dob')

    # This method is called when the age of the person needs to
    # be computed
    def _get_age(self):
        today = datetime.date.today()
        dob = self.dob
        age = today.year - dob.year
        birthday_this_year = dob.replace(year=today.year)
        if today < birthday_this_year:
            age -= 1
        return age


class Employer(Person):
    """ An employer is a person who runs a company.

    """
    # The name of the company
    company_name = Str


class Employee(Person):
    """ An employee is person with a boss and a phone number.

    """
    # The employee's boss
    boss = Instance(Employer)

    # The employee's phone number as a tuple of 3 ints
    phone = Tuple(Int, Int, Int)

    # This method is called automatically by traits to get the
    # default value for the phone number.
    def _phone_default(self):
        return (555, 555, 5555)

    # This method will be called automatically by traits when the 
    # employee's phone number changes
    def _phone_changed(self, val):
        print 'received new phone number for %s: %s' % (self.first_name, val)


if __name__ == '__main__':
    # Create an employee with a boss
    boss_john = Employer(first_name='John', last_name='Paw', 
                         company_name="Packrat's Cats")
    employee_mary = Employee(first_name='Mary', last_name='Sue', 
                             boss=boss_john)

    # Import our Enaml EmployeeView
    import enaml
    with enaml.imports():
        from employee_view import EmployeeView
    
    # Create a view and show it.
    view = EmployeeView(employee=employee_mary)
    view.show()




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from enaml import qt_toolkit


class QtTestAssistant(object):
    """ Assistant class for testing wx based components.

    This class is to be used as a mixing with the base enaml test case
    class for the components tests of the qt backend. It sets the correct
    toolkit attribute and (in the future) provide some useful methods for
    testin Qt based components.

    """

    toolkit = qt_toolkit()




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import abstractmethod

from traits.api import Bool, Instance, Str

from .container import Container, AbstractTkContainer

from ..enums import HorizontalAlign
from ..layout.geometry import Box


class AbstractTkGroupBox(AbstractTkContainer):
    """ The abstract toolkit GroupBox interface.

    """
    @abstractmethod
    def shell_title_changed(self, title):
        """ Update the title of the group box with the new value from the
        shell object.

        """
        raise NotImplementedError

    @abstractmethod
    def shell_flat_changed(self, flat):
        """ Update the flat flag of the group box with the new value from 
        the shell object.

        """
        raise NotImplementedError

    @abstractmethod
    def shell_title_align_changed(self, align):
        """ Update the title alignment to the new value from the shell 
        object.

        """
        raise NotImplementedError

    @abstractmethod
    def get_contents_margins(self):
        """ Return the Box object of margin values for the widget.

        """
        return Box(0, 0, 0, 0)


class GroupBox(Container):
    """ The GroupBox container, which introduces a group of widgets with 
    a title and usually has a border.

    """
    #: The title displayed at the top of the box.
    title = Str()

    #: The flat parameter determines if the GroupBox is displayed with 
    #: just the title and a header line (True) or with a full border 
    #: (False, the default).
    flat = Bool(False)

    #: The alignment of the title text.
    title_align = HorizontalAlign

    #: Overridden parent class trait
    abstract_obj = Instance(AbstractTkGroupBox)

    def padding_constraints(self):
        """ Overriden padding constraints method to add the contents 
        margins of the underlying group box to the specified user 
        padding.

        """
        cns = []
        margin_box = self.abstract_obj.get_contents_margins()
        user_padding = self.padding
        padding = map(sum, zip(margin_box, user_padding))
        tags = ('top', 'right', 'bottom', 'left')
        strengths = self.padding_strength
        if isinstance(strengths, basestring):
            strengths = (strengths,) * 4
        for tag, strength, padding in zip(tags, strengths, padding):
            tag = 'padding_' + tag
            sym = getattr(self, tag)
            cns.append(sym >= 0)
            if strength != 'ignore':
                cns.append((sym == padding) | strength)
        return cns




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
class ModelIndex(tuple):
    """ ModelIndexes are used to navigate an AbstractItemModel.

    A ModelIndex is a lightweight object and should be created on the
    fly, rather than attempting some form of caching. The user will
    not normally need to create indexes manually. Instead, use the
    create_index method on an AbstractItemModel.

    """
    __slots__ = ()
    
    def __new__(cls, row, column, context, model):
        """ Construct a model index.

        Arguments
        ---------
        row : int
            The row index represented by this index.

        column : int
            The column index represented by this index

        context : object
            A user supplied object that aids in navigating the user's
            model.

        model : AbstractItemModel
            The model in which this index is active. This is typically
            supplied by the create_index method of the AbstractItemModel.

        """
        return tuple.__new__(cls, (row, column, context, model))

    def __repr__(self):
        return 'ModelIndex(row=%s, col=%s, context=%s, model=%s)' % self
    
    def __str__(self):
        return self.__repr__()

    @property
    def row(self):
        return self[0]
    
    @property
    def column(self):
        return self[1]
    
    @property
    def context(self):
        return self[2]

    @property
    def model(self):
        return self[3]
        
    def parent(self):
        """ Returns the parent ModelIndex of this index.

        """
        return self.model.parent(self)

    def sibling(self, row, column):
        """ Returns the sibling ModelIndex of this index for the given
        row and column indexes.

        """
        model = self.model
        if self.row == row and self.column == column:
            return self
        return model.index(row, column, model.parent(self))

    def child(self, row, column):
        """ Returns the child ModelIndex of this index for the given
        row and column indexes.

        """
        return self.model.index(row, column, self)

    def flags(self):
        """ Returns the flags for this index.

        """
        return self.model.flags(self)


# Use the faster Cython implemented ModelIndex if available
try:
    from ..speedups.model_index import ModelIndex
except ImportError:
    pass




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from . import qt_api

if qt_api == 'pyqt':
    from PyQt4.QtCore import *

    from PyQt4.QtCore import pyqtProperty as Property
    from PyQt4.QtCore import pyqtSignal as Signal
    from PyQt4.QtCore import pyqtSlot as Slot
    from PyQt4.Qt import QCoreApplication
    from PyQt4.Qt import Qt

    __version__ = QT_VERSION_STR
    __version_info__ = tuple(map(int, QT_VERSION_STR.split('.')))

else:
    from PySide import __version__, __version_info__
    from PySide.QtCore import *



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt_test_assistant import QtTestAssistant
from .. import html

class TestQtHtml(QtTestAssistant, html.TestHtml):
    """ QtHtml tests. """

    def get_source(self, widget):
        """ Get the source of an Html widget.

        """
        return widget.toPlainText()



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import wx.calendar

from .wx_test_assistant import WXTestAssistant, skip_nonwindows
from ...components import calendar


@skip_nonwindows
class TestWXCalendar(WXTestAssistant, calendar.TestCalendar):
    """ WXCalendar tests.

    """
    def get_date(self, widget):
        """ Get a calendar's active date.

        """
        return widget.PyGetDate()

    def get_min_date(self, widget):
        """ Get a calendar's minimum date attribute.

        """
        return widget.PyGetLowerDateLimit()

    def get_max_date(self, widget):
        """ Get a calendar's maximum date attribute.

        """
        return widget.PyGetUpperDateLimit()

    def activate_date(self, widget, date):
        """ Fire an event to indicate that a date was activated.

        """
        cal_event = wx.calendar.EVT_CALENDAR
        event = wx.calendar.CalendarEvent(widget, cal_event.typeId)
        event.PySetDate(date)
        widget.GetEventHandler().ProcessEvent(event)
        self.process_wx_events(self.app)

    def select_date(self, widget, date):
        """ Fire an event to indicate that a date was selected.

        """
        cal_select_event = wx.calendar.EVT_CALENDAR_SEL_CHANGED
        event = wx.calendar.CalendarEvent(widget, cal_select_event.typeId)
        widget.PySetDate(date)
        event.PySetDate(date)
        widget.GetEventHandler().ProcessEvent(event)
        self.process_wx_events(self.app)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt_control import QtControl
from .abstract_item_model_wrapper import AbstractItemModelWrapper

from ...components.abstract_item_view import AbstractTkItemView


class QtAbstractItemView(QtControl, AbstractTkItemView):
    """ An abstract base class for implementing Qt item views.

    """
    #--------------------------------------------------------------------------
    # Setup methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Create the underlying QItemView control.

        """
        raise NotImplementedError

    def initialize(self):
        """ Initialize the widget with the attributes of this instance.

        """
        super(QtAbstractItemView, self).initialize()
        shell = self.shell_obj
        self.set_item_model(shell.item_model)

    def bind(self):
        """ Bind any event/signal handlers for the Qt Widget.

        """
        super(QtAbstractItemView, self).bind()
        widget = self.toolkit_widget
        widget.activated.connect(lambda idx: self._send_event(idx, 'activated'))
        widget.clicked.connect(lambda idx: self._send_event(idx, 'clicked'))
        widget.doubleClicked.connect(lambda idx: self._send_event(idx, 'double_clicked'))

    #--------------------------------------------------------------------------
    # Implementation
    #--------------------------------------------------------------------------
    def shell_item_model_changed(self, item_model):
        """ The change handler for the 'item_model' attribute on the 
        shell object.

        """
        self.set_item_model(item_model)

    #--------------------------------------------------------------------------
    # Widget Update Methods
    #--------------------------------------------------------------------------
    def set_item_model(self, item_model):
        """ Sets the model to use for the view.

        """
        # Save the model wrapper. PySide's object cache is sometimes unreliable
        # in various versions.
        self.model_wrapper = AbstractItemModelWrapper(item_model)
        self.widget.setModel(self.model_wrapper)

    #--------------------------------------------------------------------------
    # Event notification methods
    #--------------------------------------------------------------------------
    def _send_event(self, qindex, trait):
        """ Send a ModelIndex to the given shell trait.

        """
        model = self.model_wrapper
        index = model.from_q_index(qindex)
        getattr(self.shell_obj, trait)(index)



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from contextlib import contextmanager


class guard(object):
    """ A simple guard class that can be used to protect access 
    to arbitrary code.

    Example
    -------
    from enaml.guard import guard

    def foo():
        if not guard.guarded(bar):
            bar()
    
    def bar():
        with guard(bar):
            foo()

    """
    def __init__(self):
        self._counts = {}

    @contextmanager
    def __call__(self, *signature):
        """ Enter a guarded context for a given signature.

        Parameters
        ----------
        *signature
            Any number of positional arguments that forms the signature 
            of the guard. All of these arguments must be hashable.
        
        Returns
        -------
        result : context manager
            The 'with' statement context manager that handles the 
            lifetime of the guard.
        
        """
        counts = self._counts
        if signature in counts:
            counts[signature] += 1
        else:
            counts[signature] = 1
        try:
            yield self
        finally:
            counts[signature] -= 1
            if counts[signature] == 0:
                del counts[signature]
                
    def guarded(self, *signature):
        """ Check whether the given signature is currently guarded.

        Parameters
        ----------
        *signature
            Any number of positional arguments that forms the signature 
            of the guard. All of these arguments must be hashable.
        
        Returns
        -------
        result : bool
            Whether or not the signature is currently guarded.
        
        """
        return signature in self._counts


#: A singleton instance of the guard class. There is no need for more
#: than one.
guard = guard()





#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt import QtGui, QtCore
from .qt_bounded_date import QtBoundedDate

from ...components.date_edit import AbstractTkDateEdit
from ...guard import guard


# Workaround for an incompatibility between PySide and PyQt
try:
    qdate_to_python = QtCore.QDate.toPython
except AttributeError:
    qdate_to_python = QtCore.QDate.toPyDate


class QtDateEdit(QtBoundedDate, AbstractTkDateEdit):
    """ A Qt implementation of DateEdit.

    """
    #--------------------------------------------------------------------------
    # Setup methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the underlying QDateEdit.

        """
        self.widget = QtGui.QDateEdit(parent)

    def initialize(self):
        """ Initializes the attributes of the control.

        """
        super(QtDateEdit, self).initialize()
        self.set_format(self.shell_obj.date_format)
    
    def bind(self):
        """ Connects the signal handlers for the date edit widget.

        """
        super(QtDateEdit, self).bind()
        self.widget.dateChanged.connect(self.on_date_changed)
        
    #--------------------------------------------------------------------------
    # Implementation
    #--------------------------------------------------------------------------
    def shell_date_format_changed(self, date_format):
        """ The change handler for the 'format' attribute.

        """
        self.set_format(date_format)

    def on_date_changed(self):
        """ The signal handler for the controls's changed event.

        """
        # only emit update the shell object if the widget was 
        # changed via the ui and not programmatically.
        if not guard.guarded(self, 'updating'):
            shell = self.shell_obj
            qdate = self.widget.date()
            new_date = qdate_to_python(qdate)
            shell.date = new_date
            shell.date_changed(new_date)

    def set_date(self, date):
        """ Sets the date on the widget.

        """
        # Calling setDate will trigger the dateChanged signal.
        # We want to avoid that feeback loop since the value is
        # being set programatically.
        with guard(self, 'updating'):
            self.widget.setDate(date)

    def set_min_date(self, min_date):
        """ Sets the minimum date on the widget with the provided value.

        """
        # Calling setMinimumDate *may* trigger the dateChanged signal,
        # if the date needs to be clipped. We want to avoid that feeback 
        # loop since the value is being set programatically and the new
        # date will already have been updated by the shell object.
        with guard(self, 'updating'):
            self.widget.setMinimumDate(min_date)

    def set_max_date(self, max_date):
        """ Sets the maximum date on the widget with the provided value.

        """
        # Calling setMaximumDate *may* trigger the dateChanged signal,
        # if the date needs to be clipped. We want to avoid that feeback 
        # loop since the value is being set programatically and the new
        # date will already have been updated by the shell object.
        with guard(self, 'updating'):
            self.widget.setMaximumDate(max_date)
        
    def set_format(self, date_format):
        """ Sets the display format on the widget with the provided value.

        """
        self.widget.setDisplayFormat(date_format)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt import QtGui
from .qt_control import QtControl

from ...components.combo_box import AbstractTkComboBox
from ...guard import guard


class QtComboBox(QtControl, AbstractTkComboBox):
    """ A Qt implementation of ComboBox.

    Use a combo box to select a single item from a collection of items.

    """
    #--------------------------------------------------------------------------
    # Setup methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates a QComboBox.

        """
        self.widget = QtGui.QComboBox(parent)

    def initialize(self):
        """ Intializes the widget with the attributes of this instance.

        """
        super(QtComboBox, self).initialize()
        shell = self.shell_obj
        self.set_items(shell.labels)
        self.set_selection(shell.index)

    def bind(self):
        """ Connects the event handlers for the combo box.

        """
        super(QtComboBox, self).bind()
        self.widget.currentIndexChanged.connect(self.on_selected)

    #--------------------------------------------------------------------------
    # Implementation
    #--------------------------------------------------------------------------
    def shell_index_changed(self, index):
        """ The change handler for the 'index' attribute on the shell
        object.

        """
        self.set_selection(index)

    def shell_labels_changed(self, labels):
        """ The change handler for the 'labels' attribute on the shell
        object.

        """
        self.set_items(labels)

    def on_selected(self):
        """ The event handler for a combo box selection event.

        """
        if not guard.guarded(self, 'updating'):
            shell = self.shell_obj
            curr_index = self.widget.currentIndex()
            shell.index = curr_index

            # Only fire the selected event if we have a valid selection
            if curr_index != -1:
                shell.selected(shell.value)

    def set_items(self, str_items):
        """ Sets the items in the combo box.

        """
        # We need to avoid a feedback loop when updating the items in 
        # the combo box. Qt will emit index changed signals when the 
        # items are updated. But, the shell object has already computed
        # the proper index for the new items, so we use that to update
        # the index of the control after updating the items. The flag
        # is read by the on_selected handler to ignore updates during
        # this process.
        with guard(self, 'updating'):
            widget = self.widget
            widget.clear()
            widget.addItems(str_items)
            widget.setCurrentIndex(self.shell_obj.index)

    def set_selection(self, index):
        """ Sets the value in the combo box, or resets the combo box
        if the value is not in the list of items.

        """
        # We need to avoid a feedback loop when updating the selection
        # in the combo box. Qt will emit index changed signals when the 
        # selectino is updated. But, the shell object has already computed
        # the proper index for the new selection so we don't need to feed
        # back while doing this.
        with guard(self, 'updating'):
            self.widget.setCurrentIndex(index)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt_test_assistant import QtTestAssistant
from .. import combo_box


class TestQtComboBox(QtTestAssistant, combo_box.TestComboBox):
    """ QtComboBox tests. """

    def get_selected_text(self, widget):
        """ Get the current selected text of a combo box.

        """
        return widget.currentText()

    def get_item_text(self, widget, index):
        """ Get the text of a combo box item at a particular index.

        """
        return widget.itemText(index)

    def select_item(self, widget, index):
        """ Fire an event to simulate the selection of an item.

        """
        widget.setCurrentIndex(index)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import wx

from .wx_test_assistant import WXTestAssistant, skip_nonwindows

from .. import group_box


@skip_nonwindows
class TestGroupBox(WXTestAssistant, group_box.TestGroupBox):

    def setUp(self):
        """ Setup for GroupBox testcases.

        1. Run the parent setup.
        2. Resize the widget and process events to allow for the private
           widgets to resize.

        """
        super(TestGroupBox, self).setUp()
        window = self.component_by_name(self.view, 'win')
        abstract = window.abstract_obj
        abstract.resize((200, 200))
        self.process_wx_events(self.app)

    def get_title(self, component, widget):
        """ Returns the title text from the tookit widget

        """
        self.process_wx_events(self.app)
        abstract = component.abstract_obj
        return abstract.widget.GetTitle()

    def get_flat(self, component, widget):
        """ Returns the flat style status from the tookit widget

        """
        self.process_wx_events(self.app)
        abstract = component.abstract_obj
        return abstract.widget.GetFlat()

    def get_title_align(self, component, widget):
        """ Returns the title aligment in the tookit widget

        """
        self.process_wx_events(self.app)
        # FIXME: this check depends on the size of the current window and
        # is not very reliable.
        abstract = component.abstract_obj
        align = abstract.widget.GetAlignment()
        if align == wx.ALIGN_LEFT:
            res = 'left'
        elif align == wx.ALIGN_CENTER:
            res = 'center'
        elif align == wx.ALIGN_RIGHT:
            res = 'right'
        else:
            res = 'undefined'
        return res




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod
from collections import namedtuple, defaultdict
from contextlib import contextmanager
from weakref import ref

from traits.api import Disallow

from .byteplay import Code
from .monitors import AbstractMonitor
from .signaling import Signal
from .trait_types import UninitializedAttributeError


#------------------------------------------------------------------------------
# Expression Helpers
#------------------------------------------------------------------------------
@contextmanager
def swap_attribute(obj, attr, value):
    """ Swap an attribute of an object with the given value for the 
    duration of the context, restoring it on exit. The attribute must
    already exist on the object prior to entering the context.

    Parameters
    ----------
    obj : object
        The object which owns the attribute.
    
    attr : string
        The name of the attribute on the object.

    value : object
        The value to apply to the attribute for the duration of the
        context.
    
    """
    old = getattr(obj, attr)
    setattr(obj, attr, value)
    yield
    setattr(obj, attr, old)


#------------------------------------------------------------------------------
# Execution Scope
#------------------------------------------------------------------------------
class ExecutionScope(object):
    """ A custom mapping object that implements the scope resolution 
    order for the evaluation of code objects in Enaml expressions.

    Notes
    -----
    Strong references are kept to all objects passed to the constructor,
    so these scope objects should be created as needed and discarded in
    order to avoid issues with reference cycles.

    """
    def __init__(self, obj, identifiers, f_globals, toolkit, overrides, cb):
        """ Initialize an execution scope.

        Parameters
        ----------
        obj : BaseComponent
            The BaseComponent instance on which the expression is bound.

        identifiers : dict
            The dictionary of identifiers that are available to the
            expression. These are checked before the attribute space
            of the component.

        f_globals : dict
            The globals dict to check after checking the attribute space
            of the given component, but before checking the toolkit.

        toolkit : Toolkit
            The toolkit to check after checking the globals.

        overrides : dict
            An dictionary of override values to check before identifiers.

        cb : callable or None
            A callable which is called when an implicit attribute is 
            found and accessed on the object. The arguments passed are 
            the object and the attribute name.

        """
        self._obj = obj
        self._identifiers = identifiers
        self._f_globals = f_globals
        self._toolkit = toolkit
        self._overrides = overrides
        self._attr_cb = cb
        self._assignments = {}

    def __getitem__(self, name):
        """ Lookup an item from the namespace.

        Returns the named item from the namespace according to the
        following precedence rules:

            1) assignments
            2) override
            3) identifiers
            4) implicit attrs
            5) f_globals
            6) toolkit
            7) builtins
        
        Parameters
        ----------
        name : string
            The name that should be looked up in the namespace.

        Returns
        -------
        result : object
            The value associated with the name, if found.

        Raises
        ------
        KeyError
            If the name is not found, a KeyError is raise.

        """
        # Check the assignments dict first since this is where all
        # local variable assignments will be stored.
        dct = self._assignments
        if name in dct:
            return dct[name]

        # The overrides have the highest precedence of all framework
        # supplied values.
        dct = self._overrides
        if name in dct:
            return dct[name]
        
        # Identifiers have the highest precedence of value able to
        # be supplied by a user of the framework.
        dct = self._identifiers
        if name in dct:
            return dct[name]

        # After identifiers, the implicit attributes of the component
        # hierarchy have precedence.
        parent = self._obj
        while parent is not None:
            try:
                res = getattr(parent, name)
            except AttributeError:
                parent = parent.parent
            else:
                # Call the attribute callback if given.
                cb = self._attr_cb
                if cb is not None:
                    cb(parent, name)
                return res

        # Global variables come after implicit attributes
        dct = self._f_globals
        if name in dct:
            return dct[name]
        
        # End with the toolkit which will raise KeyError on failure.
        # Builtins will be checked by Python using the global dict.
        return self._toolkit[name]

    def __setitem__(self, name, val):
        """ Stores the value into the internal assignments dict. This 

        """
        self._assignments[name] = val

    def __delitem__(self, name):
        """ Deletes the value from the internal assignments dict.

        """
        del self._assignments[name]

    def __contains__(self, name):
        """ Return True if the name is found in the scope, False 
        otherwise.

        """
        # This method must be supplied in order for pdb to work properly
        # from within code blocks. Any attribute callback is temporarily
        # uninstalled so that it is not executed when simply checking for
        # the existance of the item in the scope.
        with swap_attribute(self, '_attr_cb', None):
            if isinstance(name, basestring):
                try:
                    self.__getitem__(name)
                except KeyError:
                    res = False
                else:
                    res = True
            else:
                res = False
        return res


#------------------------------------------------------------------------------
# Nonlocal Scope
#------------------------------------------------------------------------------
class NonlocalScope(object):
    """ An object which implements implicit attribute scoping starting
    at a given object in the tree. It is used in conjuction with a 
    nonlocals() instance to allow for explicit referencing of values
    which would otherwise be implicitly scoped.

    """
    def __init__(self, obj, cb):
        """ Initialize a nonlocal scope.

        Parameters
        ----------
        obj : BaseComponent
            The BaseComponent instance which forms the first level of
            the scope.
        
        cb : callable or None
            A callable which is called when an implicit attribute is 
            found and accessed on the object. The arguments passed are 
            the object and the attribute name.
        
        """
        self._nls_obj = obj
        self._nls_attr_cb = cb

    def __repr__(self):
        """ A pretty representation of the NonlocalScope.

        """
        templ = 'NonlocalScope[%s]'
        return templ % self._nls_obj

    def __call__(self, level=0):
        """ Returns a new nonlocal scope object offset the given number
        of levels in the hierarchy.

        Parameters
        ----------
        level : int, optional
            The number of levels up the tree to offset. The default is
            zero and indicates no offset. The level must be >= 0.

        """
        if not isinstance(level, int) or level < 0:
            msg = ('The nonlocal scope level must be an int >= 0. '
                   'Got %r instead.')
            raise ValueError(msg % level)

        offset = 0
        target = self._nls_obj
        while target is not None and offset != level:
            target = target.parent
            offset += 1

        if offset != level:
            msg = 'Scope level %s is out of range'
            raise ValueError(msg % level)

        return NonlocalScope(target, self._nls_attr_cb)

    def __getattr__(self, name):
        """ A convenience method which allows accessing items in the
        scope via getattr instead of getitem.

        """
        try:
            return self.__getitem__(name)
        except KeyError:
            msg = "%s has no attribute '%s'" % (self, name)
            raise AttributeError(msg)

    def __setattr__(self, name, value):
        """ A convenience method which allows setting items in the 
        scope via setattr instead of setitem.

        """
        if name in ('_nls_obj', '_nls_attr_cb'):
            super(NonlocalScope, self).__setattr__(name, value)
        else:
            try:
                self.__setitem__(name, value)
            except KeyError:
                msg = "%s has no attribute '%s'" % (self, name)
                raise AttributeError(msg)

    def __getitem__(self, name):
        """ Returns the named item beginning at the current scope object
        and progressing up the tree until the named attribute is found.
        A KeyError is raised if the attribute is not found.

        """
        parent = self._nls_obj
        while parent is not None:
            try:
                res = getattr(parent, name)
            except AttributeError:
                parent = parent.parent
            else:
                cb = self._nls_attr_cb
                if cb is not None:
                    cb(parent, name)
                return res
        raise KeyError(name)
    
    def __setitem__(self, name, value):
        """ Sets the value of the scope by beginning at the current scope
        object and progressing up the tree until the named attribute is 
        found. A KeyError is raise in the attribute is not found.

        """
        parent = self._nls_obj
        while parent is not None:
            # It's not sufficient to try to do setattr(...) here and 
            # catch the AttributeError, because HasStrictTraits raises
            # a TraitError in these cases and it becomes impossible
            # to distinguish that error from a trait typing error 
            # without checking the message of the exception.
            try:
                getattr(parent, name)
            except UninitializedAttributeError:
                pass
            except AttributeError:
                parent = parent.parent
                continue
            setattr(parent, name, value)
            return
        raise KeyError(name)

    def __contains__(self, name):
        """ Return True if the name is found in the scope, False 
        otherwise.

        """
        with swap_attribute(self, '_nls_attr_cb', None):
            if isinstance(name, basestring):
                try:
                    self.__getitem__(name)
                except KeyError:
                    res = False
                else:
                    res = True
            else:
                res = False
        return res


#------------------------------------------------------------------------------
# Abstract Expression
#------------------------------------------------------------------------------
class AbstractExpression(object):
    """ The base abstract expression class which defines the base api
    for Expression handlers. These objects are typically created by
    the Enaml operators.

    """
    __metaclass__ = ABCMeta

    #: A signal which is emitted when the expression has changed. It is
    #: emmitted with three arguments: expression, name, and value; where 
    #: expression is the instance which emitted the signal, name is the 
    #: attribute name to which the expression is bound, and value is the
    #: computed value of the expression.
    expression_changed = Signal()

    def __init__(self, obj, name, code, identifiers, f_globals, toolkit):
        """ Initializes and expression object.

        Parameters
        ----------
        obj : BaseComponent
            The base component to which this expression is being bound.

        name : string
            The name of the attribute on the owner to which this 
            expression is bound.

        code : types.CodeType object
            The compiled code object for the Python expression.

        identifiers : dict
            The dictionary of identifiers that are available to the
            expression.

        f_globals : dict
            The globals dictionary in which the expression should execute.

        toolkit : Toolkit
            The toolkit that was used to create the object and in which
            the expression should execute.

        """
        self.obj_ref = ref(obj)
        self.name = name
        self.code = code
        self.identifiers = identifiers
        self.f_globals = f_globals
        self.toolkit = toolkit

    @abstractmethod
    def eval(self):
        """ Evaluates the expression and returns the result. If an 
        expression does not provide (or cannot provide) a value, it 
        should return NotImplemented.
        
        Returns
        -------
        result : object or NotImplemented
            The result of evaluating the expression or NotImplemented
            if the expression is unable to provide a value.
        
        """
        raise NotImplementedError

    @abstractmethod
    def notify(self, old, new):
        """ A method called by the owner component when the trait on 
        which it is bound has changed.

        Parameters
        ----------
        old : object
            The old value of the attribute.
        
        new : object
            The new value of the attribute.
    
        """
        raise NotImplementedError


#------------------------------------------------------------------------------
# Simple Expression
#------------------------------------------------------------------------------
class SimpleExpression(AbstractExpression):
    """ A concrete implementation of AbstractEvalExpression. An instance
    of SimpleExpression does not track changes in the expression or emit
    the expression_changed signal. Is also does not support notification.

    """
    def eval(self):
        """ Evaluates and returns the results of the expression.

        """
        obj = self.obj_ref()
        if obj is None:
            return NotImplemented
        
        overrides = {'nonlocals': NonlocalScope(obj, None)}
        identifiers = self.identifiers
        f_globals = self.f_globals
        toolkit = self.toolkit
        scope = ExecutionScope(
            obj, identifiers, f_globals, toolkit, overrides, None,
        )

        with toolkit:
            res =  eval(self.code, f_globals, scope)
        
        return res

    def notify(self, old, new):
        """ A no-op notification method since SimpleExpression does not
        support notification.

        """
        pass


#------------------------------------------------------------------------------
# Notifification Expression
#------------------------------------------------------------------------------
class NotificationExpression(AbstractExpression):
    """ A concrete implementation of AbstractExpression. An instance 
    of NotificationExpression does not support evaluation and does not
    emit the expression_changed signal, but it does support notification. 
    An 'event' object will be added to the scope of the expression which
    will contain information about the trait change.

    """
    #: A namedtuple which is used to pass arguments to the expression.
    event = namedtuple('event', 'obj name old new')

    def eval(self):
        """ A no-op eval method since NotificationExpression does not
        support evaluation.

        """
        return NotImplemented

    def notify(self, old, new):
        """ Evaluates the underlying expression, and provides an 'event'
        object in the evaluation scope which contains information about 
        the trait change.

        """
        obj = self.obj_ref()
        if obj is None:
            return

        identifiers = self.identifiers
        f_globals = self.f_globals
        toolkit = self.toolkit
        override = {
            'event': self.event(obj, self.name, old, new),
            'nonlocals': NonlocalScope(obj, None),
        }
        scope = ExecutionScope(
            obj, identifiers, f_globals, toolkit, override, None,
        )

        with toolkit:
            eval(self.code, f_globals, scope)


#------------------------------------------------------------------------------
# Update Expression
#------------------------------------------------------------------------------
class UpdateExpression(AbstractExpression):
    """ A concrete implementation of AbstractExpression which sets the 
    value on the contituents of the expression according to any provided
    inverters.

    """
    def __init__(self, inverter_classes, *args):
        """ Initialize an UpdateExpression

        Parameters
        ----------
        inverter_classes : iterable of AbstractInverter subclasses
            An iterable of concrete AbstractInverter subclasses which
            will invert the given expression code into a mirrored 
            operation which sets the value on the appropriate object.
        
        *args
            The arguments required to initialize an AbstractExpression
        
        """
        super(UpdateExpression, self).__init__(*args)
        
        inverters = []
        bp_code = Code.from_code(self.code)
        code_list = bp_code.code

        for inv_cls in inverter_classes:
            inverter = inv_cls()
            new_code = inverter.get_inverted_code(code_list)
            if new_code is not None:
                bp_code.code = new_code
                inverters.append(bp_code.to_code())
        
        if not inverters:
            msg = ("Unable to delegate expression to the '%s' attribute of "
                   "the %s object. The provided expression is not structured "
                   "in a way which is suitable for delegation by any of "
                   "the supplied code inverters.")
            raise ValueError(msg % (self.name, self.obj_ref()))

        self.inverters = tuple(inverters)
    
    def eval(self):
        """ A no-op eval method since UpdateExpression does not support 
        evaluation.

        """
        return NotImplemented

    def notify(self, old, new):
        """ A notification method which runs through the list of inverted
        code objects which attempt to set the value. The process stops on
        the first successful inversion. If none of the invertors are 
        successful, a RuntimeError is raised.

        """
        obj = self.obj_ref()
        if obj is None:
            return

        # The override dict is populated with information about the
        # expression and the change. The values allow the generated
        # inverter codes to access the information which is required
        # to perform the operation. The items are added using names
        # which are not valid Python identifiers and therefore do
        # not risk clashing with names in the expression. This is
        # the same technique used by the Python interpreter itself.
        identifiers = self.identifiers
        f_globals = self.f_globals
        toolkit = self.toolkit
        overrides = {
            '_[expr]': self, '_[obj]': obj, '_[old]': old, '_[new]': new, 
            '_[name]': self.name, 'nonlocals': NonlocalScope(obj, None),
        }
        scope = ExecutionScope(
            obj, identifiers, f_globals, toolkit, overrides, None,
        )

        # Run through the inverters, giving each a chance to do the
        # inversion. The process ends with the first success. If 
        # none of the invertors are successful an error is raised.
        with toolkit:
            for inverter in self.inverters:
                if eval(inverter, f_globals, scope):
                    break
            else:
                msg = ("Unable to delegate expression to the %r attribute "
                       "of the %s object. None of the provided inverters were "
                       "successful in assigning the value.")
                raise RuntimeError(msg % (self.name, obj))


#------------------------------------------------------------------------------
# Subscription Expression
#------------------------------------------------------------------------------
class _ImplicitAttributeBinder(object):
    """ A thin class which supports attaching a notifier to an implicit
    attribute lookup.

    """
    # This doesn't need to be provided as a monitor because implicit 
    # attribute lookups, when successful, will always be on an instance
    # of BaseComponent and should never need to be hooked by an Enaml 
    # extension.
    def __init__(self, parent):
        """ Initialize an _ImplicitAttributeBinder

        Parameters
        ----------
        parent : SubscriptionExpression
            The parent SubscriptionExpression instance. Only a weak
            reference to the parent is stored.

        """
        self.parent_ref = ref(parent)
    
    def __call__(self, obj, name):
        """ Binds the change handler to the given object/attribute
        pair, provided the attribute points to a valid trait.

        Parameters
        ----------
        obj : BaseComponent
            The BaseComponent instance that owns the attribute.
        
        name : string
            The attribute name of interest
             
        """
        trait = obj.trait(name)
        if trait is not None and trait is not Disallow:
            obj.on_trait_change(self.notify, name)
    
    def notify(self):
        """ The trait change handler callback. It calls the monitor
        changed method on the parent when the trait changes, provided 
        the parent has not already been garbage collected.
        
        """
        parent = self.parent_ref()
        if parent is not None:
            parent._on_monitor_changed()


class SubscriptionExpression(AbstractExpression):
    """ A concrete implementation of AbstractExpression. An instance 
    of SubcriptionExpression emits the expression_changed signal when
    the value of the underlying expression changes. It does not 
    support notification.

    """
    def __init__(self, monitor_classes, *args):
        """ Initialize a SubscriptionExpression

        Parameters
        ----------
        monitor_classes : iterable of AbstractMonitor subclasses
            An iterable of AbstractMonitor subclasses which will
            be used to generating the monitoring code for the
            expression.
        
        *args
            The arguments required to initialize an AbstractExpression
        
        """
        super(SubscriptionExpression, self).__init__(*args)
        
        # Create the monitor instances and connect their signals
        monitors = []
        handler = self._on_monitor_changed
        for mcls in monitor_classes:
            if not issubclass(mcls, AbstractMonitor):
                msg = ('Monitors must be subclasses of AbstractMonitor. '
                       'Got %s instead.')
                raise TypeError(msg % mcls)
            monitor = mcls()
            monitor.expression_changed.connect(handler)
            monitors.append(monitor)

        # Collect the generated code from the monitors that will be
        # inserted into the code for the expression.
        bp_code = Code.from_code(self.code)
        code_list = list(bp_code.code)
        insertions = defaultdict(list)
        for monitor in monitors:
            for idx, code in monitor.get_insertion_code(code_list):
                insertions[idx].extend(code)
        
        # Create a new code list which interleaves the code generated
        # by the monitors at the appropriate location in the expression.
        new_code = []
        for idx, code_op in enumerate(code_list):
            if idx in insertions:
                new_code.extend(insertions[idx])
            new_code.append(code_op)
        
        bp_code.code = new_code
        self.eval_code = bp_code.to_code()
        self.monitors = tuple(monitors)
        self.implicit_binder = _ImplicitAttributeBinder(self)
        self.old_value = NotImplemented

    def _on_monitor_changed(self):
        """ The signal callback which is fired from a monitor when the
        expression changes. It will fire the expression_changed signal
        provided that the value of the expression has actually changed.

        """
        new_value = self.eval()
        
        # Guard against exceptions being raise during comparison, such 
        # as when comparing two numpy arrays.
        try:
            different = new_value != self.old_value
        except Exception:
            different = True

        if different:
            self.old_value = new_value
            self.expression_changed(self, self.name, new_value)

    def eval(self):
        """ Evaluates the expression and returns the result. It also
        resets the monitors before evaluating the expression to help
        ensures that duplicate notifications are avoided.

        """
        # Reset the monitors before every evaluation so that any old 
        # notifiers are disconnected. This avoids muti-notifications.
        self.implicit_binder = binder = _ImplicitAttributeBinder(self)
        for monitor in self.monitors:
            monitor.reset()

        obj = self.obj_ref()
        if obj is None:
            return NotImplemented
        
        identifiers = self.identifiers
        f_globals = self.f_globals
        toolkit = self.toolkit
        overrides = {'nonlocals': NonlocalScope(obj, binder)}
        scope = ExecutionScope(
            obj, identifiers, f_globals, toolkit, overrides, binder,
        )

        with toolkit:
            res = eval(self.eval_code, f_globals, scope)

        return res

    def notify(self, old, new):
        """ A no-op notification method since SubscriptionExpression does
        not support notification.

        """
        pass


#------------------------------------------------------------------------------
# Delegation Expression
#------------------------------------------------------------------------------
class DelegationExpression(SubscriptionExpression):
    """ A SubscriptionExpression subclass that adds notification support
    by setting the value on the contituents of the expression according
    to any provide inverters.

    """
    def __init__(self, inverter_classes, *args):
        """ Initialize a DelegationExpression

        Parameters
        ----------
        inverter_classes : iterable of AbstractInverter subclasses
            An iterable of concrete AbstractInverter subclasses which
            will invert the given expression code into a mirrored 
            operation which sets the value on the appropriate object.
        
        *args
            The arguments required to initialize a SubscriptionExpression
        
        """
        super(DelegationExpression, self).__init__(*args)
        
        inverters = []
        bp_code = Code.from_code(self.code)
        code_list = bp_code.code

        for inv_cls in inverter_classes:
            inverter = inv_cls()
            new_code = inverter.get_inverted_code(code_list)
            if new_code is not None:
                bp_code.code = new_code
                inverters.append(bp_code.to_code())
        
        if not inverters:
            msg = ("Unable to delegate expression to the '%s' attribute of "
                   "the %s object. The provided expression is not structured "
                   "in a way which is suitable for delegation by any of "
                   "the supplied code inverters.")
            raise ValueError(msg % (self.name, self.obj_ref()))

        self.inverters = inverters
    
    def notify(self, old, new):
        """ A notification method which runs through the list of inverted
        code objects which attempt to set the value. The process stops on
        the first successful inversion. If none of the invertors are 
        successful, a RuntimeError is raised.

        """
        obj = self.obj_ref()
        if obj is None:
            return
        
        # We don't need to attempt the inversion if the new value
        # is the same as the last value generated by the expression.
        # This helps prevent bouncing back and forth which can be
        # induced by excessive notification. Guard against exceptions
        # being raise during comparison, such as when comparing two
        # numpy arrays.
        try:
            different = new != self.old_value
        except Exception:
            different = True

        if not different:
            return

        # The override dict is populated with information about the
        # expression and the change. The values allow the generated
        # inverter codes to access the information which is required
        # to perform the operation. The items are added using names
        # which are not valid Python identifiers and therefore do
        # not risk clashing with names in the expression. This is
        # the same technique used by the Python interpreter itself.
        identifiers = self.identifiers
        f_globals = self.f_globals
        toolkit = self.toolkit
        overrides = {
            '_[expr]': self, '_[obj]': obj, '_[old]': old, '_[new]': new, 
            '_[name]': self.name, 'nonlocals': NonlocalScope(obj, None),
        }
        scope = ExecutionScope(
            obj, identifiers, f_globals, toolkit, overrides, None,
        )

        # Run through the inverters, giving each a chance to do the
        # inversion. The process ends with the first success. If 
        # none of the invertors are successful an error is raised.
        with toolkit:
            for inverter in self.inverters:
                if eval(inverter, f_globals, scope):
                    break
            else:
                msg = ("Unable to delegate expression to the '%s' attribute "
                       "of the %s object. None of the provided inverters were "
                       "successful in assigning the value.")
                raise RuntimeError(msg)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod

from traits.api import HasTraits, Range

from casuarius import ConstraintVariable, LinearSymbolic, Strength, STRENGTH_MAP

from .constrainable import Constrainable, PaddingConstraints
from .geometry import Box

from ..auxiliary.decorators import deprecated
from ..core.trait_types import CoercingInstance


#------------------------------------------------------------------------------
# Default Spacing
#------------------------------------------------------------------------------
class DefaultSpacing(HasTraits):
    """ A class which encapsulates the default spacing parameters for
    the various layout helper objects.

    """
    #: The space between abutted components
    ABUTMENT = Range(low=0, value=10)

    #: The space between aligned anchors
    ALIGNMENT = Range(low=0, value=0)

    #: The margins for box helpers
    BOX_MARGINS = CoercingInstance(Box, default=Box(0, 0, 0, 0))


# We only require a singleton of DefaultSpacing
DefaultSpacing = DefaultSpacing()


#------------------------------------------------------------------------------
# Abstract Constrainable Interface
#------------------------------------------------------------------------------
class ABConstrainable(object):
    """ Abstract base class for objects that can be laid out by these 
    helpers. Minimally, they need to have `top`, `bottom`, `left`, `right`
    attributes that are `LinearSymbolic` instances.

    """
    __metaclass__ = ABCMeta


ABConstrainable.register(Constrainable)


#------------------------------------------------------------------------------
# Helper Functions
#------------------------------------------------------------------------------
def expand_constraints(component, constraints):
    """ A function which expands any DeferredConstraints in the provided
    list. This is a generator function which yields the flattened stream 
    of constraints.

    Paramters
    ---------
    component : Constrainable
        The constrainable component with which the constraints are 
        associated. This will be passed to the .get_constraints() 
        method of any DeferredConstraint instance.
    
    constraints : list
        The list of constraints.
    
    Yields
    ------
    constraints
        The stream of expanded constraints.

    """
    for cn in constraints:
        if isinstance(cn, DeferredConstraints):
            for item in cn.get_constraints(component):
                if item is not None:
                    yield item
        else:
            if cn is not None:
                yield cn


def is_spacer(item):
    """ Returns True if the given item can be considered a spacer, False
    other otherwise.

    """
    return isinstance(item, (Spacer, int))


def is_really_visible(item):
    """ Returns True if the given item is actually visible on the screen.

    This is determined by checking all the ancestors of the component
    for visibility and only return True if they are all visible.

    """
    if not item.visible:
        return False
    for ancestor in item.traverse_ancestors():
        if not getattr(ancestor, 'visible', True):
            return False
    return True


def clear_invisible(items):
    """ Take a list of Components and other layout items and remove 
    logically invisible Components.

    This takes into account redundant spacer objects that may surround 
    invisible objects. Spacer objects that appear before an invisible 
    Component will be removed.

    Lists that consist solely of spacers will result in an empty list.

    Parameters
    ----------
    items : list
        The list of layout items to filter.
    
    Returns
    -------
    results : list
        A new list with logically invisible items removed.

    """
    vis = []
    push = vis.append
    pop = vis.pop
    for item in items:
        if isinstance(item, Constrainable) and not is_really_visible(item):
            if len(vis) > 0 and is_spacer(vis[-1]):
                pop()
        else:
            push(item)
    if all(is_spacer(item) for item in vis):
        vis = []
    return vis


#------------------------------------------------------------------------------
# Deferred Constraints
#------------------------------------------------------------------------------
class DeferredConstraints(object):
    """ Abstract base class for objects that will yield lists of 
    constraints upon request.

    """
    __metaclass__ = ABCMeta

    def __init__(self, *args, **kwds):
        """ Initialize a DeferredConstraints instance.

        """
        self.clear_invisible = kwds.get('clear_invisible', True)
        # __or__() will set these default strength and weight. If
        # provided, they will be combined with the constraints created
        # by this instance.
        self.default_strength = None
        self.default_weight = None

    def __or__(self, other):
        """ Set the strength of all of the constraints to a common 
        strength.

        """
        if isinstance(other, (float, int, long)):
            self.default_weight = float(other)
        elif isinstance(other, basestring):
            if other not in STRENGTH_MAP:
                raise ValueError('Invalid strength %r' % other)
            self.default_strength = STRENGTH_MAP[other]
        elif isinstance(other, Strength):
            self.default_strength = other
        else:
            msg = 'Strength must be a string. Got %s instead.'
            raise TypeError(msg % type(other))
        return self

    def when(self, switch):
        """ A simple method that can be used to switch off the generated
        constraints depending on a boolean value.

        """
        if switch:
            return self
    
    def get_constraints(self, component):
        """ Returns a list of weighted LinearConstraints.

        Parameters
        ----------
        component : Component or None
            The component that owns this DeferredConstraints. It can 
            be None for contexts in which there is not a containing
            component, such as in certain nested DeferredConstraints.

        Returns
        -------
        result : list of LinearConstraints
            The list of LinearConstraint objects which have been 
            weighted by any provided strengths and weights.
        
        """
        cn_list = self._get_constraints(component)
        strength = self.default_strength
        if strength is not None:
            cn_list = [cn | strength for cn in cn_list]
        weight = self.default_weight
        if weight is not None:
            cn_list = [cn | weight for cn in cn_list]
        return cn_list

    @abstractmethod
    def _get_constraints(self, component):
        """ Returns a list of LinearConstraint objects.

        Subclasses must implement this method to actually yield their
        constraints. Users of instances should instead call the 
        `get_weighted_constraints()` method which will combine these 
        constraints with the `default_strength` and/or the 
        `default_weight` if one or both are provided.

        Parameters
        ----------
        component : Component or None
            The component that owns this DeferredConstraints. It can 
            be None for contexts in which there is not a containing
            component, such as in certain nested DeferredConstraints.

        Returns
        -------
        result : list of LinearConstraints
            The list of LinearConstraint objects for this deferred 
            instance.

        """
        raise NotImplementedError


#------------------------------------------------------------------------------
# Deferred Constraints Implementations
#------------------------------------------------------------------------------
class DeferredConstraintsFunction(DeferredConstraints):
    """ A concrete implementation of DeferredConstraints which will 
    call a function to get the constraint list upon request.

    """
    def __init__(self, func, *args, **kwds):
        """ Initialize a DeferredConstraintsFunction.

        Parameters
        ----------
        func : callable
            A callable object which will return the list of constraints.
        
        *args
            The arguments to pass to 'func'.
        
        **kwds
            The keyword arguments to pass to 'func'.
        
        """
        super(DeferredConstraintsFunction, self).__init__(func, *args, **kwds)
        self.func = func
        self.args = args
        self.kwds = kwds

    def _get_constraints(self, component):
        """ Abstract method implementation which calls the underlying 
        function to generate the list of constraints.

        """
        return self.func(*self.args, **self.kwds)
    

class AbutmentHelper(DeferredConstraints):
    """ A concrete implementation of DeferredConstraints which will
    lay out its components by abutting them in a given orientation.

    """
    def __init__(self, *items, **metadata):
        """ Initialize an AbutmentHelper.

        Parameters
        ----------
        *items
            The components to abut in the given orientation.

        **metadata
            Metadata related to how the abutment should behave. The
            following options are currently supported:
                
                orientation
                    A string which is either 'horizontal' or 'vertical'
                    which indicates the abutment orientation. The 
                    default is 'vertical'.
                
                spacing
                    An integer >= 0 which indicates how many pixels of
                    inter-element spacing to use during abutment. The 
                    default is the value of DefaultSpacing.ABUTMENT.

        """
        super(AbutmentHelper, self).__init__(*items, **metadata)
        self.items = items
        self.orientation = metadata.get('orientation', 'vertical')
        self.spacing = metadata.get('spacing', DefaultSpacing.ABUTMENT)

    def __repr__(self):
        items = ', '.join(map(repr, self.items))
        return '{0}({1})'.format(self.orientation, items)

    def _get_constraints(self, component):
        """ Abstract method implementation which applies the constraints
        to the given items, after filtering them for None values.

        """
        items = [item for item in self.items if item is not None]
        if self.clear_invisible:
            items = clear_invisible(items)
        cns = []
        factories = AbutmentConstraintFactory.from_items(
            items, self.orientation, self.spacing,
        )
        for f in factories:
            cns.extend(f.constraints())
        return cns


class AlignmentHelper(DeferredConstraints):
    """ A deferred constraints helper class that lays out with a given
    anchor to align.

    """
    def __init__(self, *items, **metadata):
        """ Initialize an AlignmentHelper.

        Parameters
        ----------
        *items
            The components to align on the given anchor.

        **metadata
            Metadata related to how the abutment should behave. The
            following options are currently supported:
                
                anchor
                    A string which is either 'left', 'right', 'top', 
                    'bottom', 'v_center', or 'h_center'. The default
                    is 'v_center'.
                
                spacing
                    An integer >= 0 which indicates how many pixels of
                    inter-element spacing to use during alignement. The 
                    default is the value of DefaultSpacing.ALIGNMENT.

        """
        super(AlignmentHelper, self).__init__(*items, **metadata)
        self.items = items
        self.anchor = metadata.get('anchor', 'v_center')
        self.spacing = metadata.get('spacing', DefaultSpacing.ALIGNMENT)

    def __repr__(self):
        items = ', '.join(map(repr, self.items))
        return 'align({0!r}, {1})'.format(self.anchor, items)

    def _get_constraints(self, component):
        """ Abstract method implementation which applies the constraints
        to the given items, after filtering them for None values.

        """
        items = [item for item in self.items if item is not None]
        if self.clear_invisible:
            items = clear_invisible(items)
        cns = []
        factories = AlignmentConstraintFactory.from_items(
            items, self.anchor, self.spacing,
        )
        for f in factories:
            cns.extend(f.constraints())
        return cns

    
class LinearBoxHelper(DeferredConstraints):
    """ A DeferredConstraints helper class that lays out Components 
    either horizontally or vertically and taking up all of the space 
    in the orthogonal direction. Instances of this class are nestable.

    """
    #: A mapping orientation to the anchor names needed to make the
    #: constraints on the containing component.
    orientation_map = {
        'horizontal': ('left', 'right'),
        'vertical': ('top', 'bottom'),
    }

    def __init__(self, *items, **metadata):
        """ Initialize an AlignmentHelper.

        Parameters
        ----------
        *items
            The components to align on the given anchor.

        **metadata
            Metadata related to how the abutment should behave. The
            following options are currently supported:
                
                orientation
                    A string which is either 'horizontal' or 'vertical'
                    which indicates the abutment orientation. The 
                    default is 'vertical'.
                
                spacing
                    An integer >= 0 which indicates how many pixels of
                    inter-element spacing to use during abutment. The 
                    default is the value of DefaultSpacing.ABUTMENT.
                
                margin
                    A int, tuple of ints, or Box of ints >= 0 which 
                    indicate how many pixels of margin to add around 
                    the bounds of the box. The default is the value of 
                    DefaultSpacing.BOX_MARGIN.
                
        """
        super(LinearBoxHelper, self).__init__(*items, **metadata)

        orientation = metadata.get('orientation', 'vertical')
        if orientation == 'horizontal':
            ortho_orientation = 'vertical'
        elif orientation == 'vertical':
            ortho_orientation = 'horizontal'
        else:
            msg = ("Expected 'horizontal' or 'vertical' for orientation. "
                   "Got {!r} instead.")
            raise ValueError(msg.format(orientation))
        
        self.items = items
        self.orientation = orientation
        self.ortho_orientation = ortho_orientation
        self.spacing = metadata.get('spacing', DefaultSpacing.ABUTMENT)
        self.margins = Box(metadata.get('margins', DefaultSpacing.BOX_MARGINS))
        
        for attr in ('top', 'bottom', 'left', 'right'):
            label = '{0}_{1:x}'.format(self.orientation[0]+'box', id(self))
            var = ConstraintVariable('{0}_{1}'.format(attr, label))
            setattr(self, attr, var)
    
    def __repr__(self):
        items = ', '.join(map(repr, self.items))
        return '{0}box({1})'.format(self.orientation[0], items)

    def _get_constraints(self, component):
        """ Abstract method implementation which applies the constraints
        to the given items, after filtering them for None values.

        """
        items = [item for item in self.items if item is not None]
        if self.clear_invisible:
            items = clear_invisible(items)

        if len(items) == 0:
            return []
        
        get = getattr
        first_name, last_name = self.orientation_map[self.orientation]
        first_boundary = get(self, first_name)
        last_boundary = get(self, last_name)
        first_name, last_name = self.orientation_map[self.ortho_orientation]
        first_ortho_boundary = get(self, first_name)
        last_ortho_boundary = get(self, last_name)

        # Setup the initial outer constraints of the box
        if component is not None:
            # This box helper is inside a real component, not just nested 
            # inside of another box helper. Check if the component is a
            # PaddingConstraints object and use it's contents anchors.
            attrs = ['top', 'bottom', 'left', 'right']
            if isinstance(component, PaddingConstraints):
                other_attrs = ['contents_' + attr for attr in attrs]
            else:
                other_attrs = attrs[:]
            constraints = [
                get(self, attr) == get(component, other) 
                for (attr, other) in zip(attrs, other_attrs)
            ]
        else:
            constraints = []
        
        # Create the margin spacers that will be used.
        margins = self.margins
        if self.orientation == 'vertical':
            first_spacer = EqSpacer(margins.top)
            last_spacer = EqSpacer(margins.bottom)
            first_ortho_spacer = FlexSpacer(margins.left)
            last_ortho_spacer = FlexSpacer(margins.right)
        else:
            first_spacer = EqSpacer(margins.left)
            last_spacer = EqSpacer(margins.right)
            first_ortho_spacer = FlexSpacer(margins.top)
            last_ortho_spacer = FlexSpacer(margins.bottom)

        # Add a pre and post padding spacer if the user hasn't specified 
        # their own spacer as the first/last element of the box items.
        if not is_spacer(items[0]):
            pre_along_args = [first_boundary, first_spacer]
        else:
            pre_along_args = [first_boundary]
        if not is_spacer(items[-1]):
            post_along_args = [last_spacer, last_boundary]
        else:
            post_along_args = [last_boundary]

        # Accummulate the constraints in the direction of the layout
        along_args = pre_along_args + items + post_along_args
        kwds = dict(
            orientation=self.orientation, spacing=self.spacing,
            clear_invisible=self.clear_invisible,
        )
        helpers = [AbutmentHelper(*along_args, **kwds)]
        kwds['orientation'] = self.ortho_orientation
        for item in items:
            # Add the helpers for the ortho constraints
            if isinstance(item, ABConstrainable):
                items = (
                    first_ortho_boundary, first_ortho_spacer,
                    item, last_ortho_spacer, last_ortho_boundary,
                )
                helpers.append(AbutmentHelper(*items, **kwds)) 
            # Pull out nested helpers so that their constraints get
            # generated during the pass over the helpers list.
            if isinstance(item, DeferredConstraints):
                helpers.append(item)
        
        # Pass over the list of child helpers and generate the 
        # flattened list of constraints.
        for helper in helpers:
            constraints.extend(helper.get_constraints(None))

        return constraints


ABConstrainable.register(LinearBoxHelper)


#------------------------------------------------------------------------------
# Abstract Constraint Factory
#------------------------------------------------------------------------------
class AbstractConstraintFactory(object):
    """ An abstract constraint factory class. Subclasses must implement
    the 'constraints' method implement which returns a LinearConstraint
    instance.

    """
    __metaclass__ = ABCMeta

    @staticmethod
    def validate(items):
        """ A validator staticmethod that insures a sequence of items is 
        appropriate for generating a sequence of linear constraints. The 
        following conditions are verified of the sequence of given items:

            * The number of items in the sequence is 0 or >= 2.

            * The first and last items are instances of either
              LinearSymbolic or Constrainable.

            * All of the items in the sequence are instances of 
              LinearSymbolic, Constrainable, Spacer, or int.
    
        If any of the above conditions do not hold, an exception is 
        raised with a (hopefully) useful error message.

        """
        if len(items) == 0:
            return

        if len(items) < 2:
            msg = 'Two or more items required to setup abutment constraints.'
            raise ValueError(msg)
        
        extrema_types = (LinearSymbolic, ABConstrainable)
        def extrema_test(item):
            return isinstance(item, extrema_types)
        
        item_types = (LinearSymbolic, ABConstrainable, Spacer, int)
        def item_test(item):
            return isinstance(item, item_types)

        if not all(extrema_test(item) for item in (items[0], items[-1])):
            msg = ('The first and last items of a constraint sequence '
                   'must be anchors or Components. Got %s instead.')
            args = [type(items[0]), type(items[1])]
            raise TypeError(msg % args)
        
        if not all(map(item_test, items)):
            msg = ('The allowed items for a constraint sequence are'
                   'anchors, Components, Spacers, and ints. '
                   'Got %s instead.')
            args = [type(item) for item in items]
            raise TypeError(msg % args)

    @abstractmethod
    def constraints(self):
        """ An abstract method which must be implemented by subclasses.
        It should return a list of LinearConstraint instances.

        """
        raise NotImplementedError


#------------------------------------------------------------------------------
# Abstract Constraint Factory Implementations
#------------------------------------------------------------------------------
class BaseConstraintFactory(AbstractConstraintFactory):
    """ A base constraint factory class that implements basic common
    logic. It is not meant to be used directly but should rather be 
    subclassed to be useful.

    """
    def __init__(self, first_anchor, spacer, second_anchor):
        """ Create an base constraint instance.

        Parameters
        ----------
        first_anchor : LinearSymbolic
            A symbolic object that can be used in a constraint expression.
        
        spacer : Spacer
            A spacer instance to put space between the items.
        
        second_anchor : LinearSymbolic
            The second anchor for the constraint expression.

        """
        self.first_anchor = first_anchor
        self.spacer = spacer
        self.second_anchor = second_anchor

    def constraints(self):
        """ Returns LinearConstraint instance which is formed through
        an appropriate linear expression for the given space between 
        the anchors.

        """
        first = self.first_anchor
        second = self.second_anchor
        spacer = self.spacer
        return spacer.constrain(first, second)


class SequenceConstraintFactory(BaseConstraintFactory):
    """ A BaseConstraintFactory subclass that represents a constraint
    between two anchors of different components separated by some amount
    of space. It has a '_make_cns' classmethod which will create a list
    of constraint factory instances from a sequence of items, the two
    anchor names, and a default spacing.

    """
    @classmethod
    def _make_cns(cls, items, first_anchor_name, second_anchor_name, spacing):
        """ A classmethod that generates a list of constraints factories
        given a sequence of items, two anchor names, and default spacing. 

        Parameters
        ----------
        items : sequence
            A valid sequence of constrainable objects. These inclue
            instances of Constrainable, LinearSymbolic, Spacer,
            and int.
        
        first_anchor_name : string
            The name of the anchor on the first item in a constraint
            pair.
        
        second_anchor_name : string
            The name of the anchor on the second item in a constraint
            pair.
        
        spacing : int
            The spacing to use between items if no spacing is explicitly 
            provided by in the sequence of items.
        
        Returns
        -------
        result : list
            A list of constraint factory instance.
        
        """
        # Make sure the items we'll be dealing with are valid for the
        # algorithm. This is a basic validation. Further error handling
        # is performed as needed.
        cls.validate(items)
    
        # The list of constraints we'll be creating for the given 
        # sequence of items.
        cns = []

        # The list of items is treated as a stack. So we want to first
        # reverse it so the first items are at the top of the stack.
        items = list(reversed(items))

        while items:

            # Grab the item that will provide the first anchor
            first_item = items.pop()

            # first_item will be a Constrainable or a LinearSymbolic.
            # For the first iteration, this is enforced by 'validate'.
            # For subsequent iterations, this condition is enforced by 
            # the fact that this loop only pushes those types back onto 
            # the stack.
            if isinstance(first_item, ABConstrainable):
                first_anchor = getattr(first_item, first_anchor_name)
            elif isinstance(first_item, LinearSymbolic):
                first_anchor = first_item
            else:
                raise TypeError('This should never happen')
            
            # Grab the next item off the stack. It will be an instance
            # of Constrainable, LinearSymbolic, Spacer, or int. If it
            # can't provide an anchor, we grab the item after it which
            # *should* be able to provide one. If no space is given, we 
            # use the provided default space.
            next_item = items.pop()
            if isinstance(next_item, Spacer):
                spacer = next_item
                second_item = items.pop()
            elif isinstance(next_item, int):
                spacer = EqSpacer(next_item)
                second_item = items.pop()
            elif isinstance(next_item, (ABConstrainable, LinearSymbolic)):
                spacer = EqSpacer(spacing)
                second_item = next_item
            else:
                raise ValueError('This should never happen')
            
            # If the second_item can't provide an anchor, such as two 
            # spacers next to each other, then this is an error and we
            # raise an appropriate exception.
            if isinstance(second_item, ABConstrainable):
                second_anchor = getattr(second_item, second_anchor_name)   
            elif isinstance(second_item, LinearSymbolic):
                second_anchor = second_item
            else:
                msg = 'Expected anchor or Constrainable. Got %r instead.'
                raise TypeError(msg % second_item)
            
            # Create the class instance for this constraint
            factory = cls(first_anchor, spacer, second_anchor)

            # If there are still items on the stack, then the second_item
            # will be used as the first_item in the next iteration. 
            # Otherwise, we have exhausted all constraints and can exit.
            if items:
                items.append(second_item)

            # Finally, store away the created factory for returning.
            cns.append(factory)

        return cns


class AbutmentConstraintFactory(SequenceConstraintFactory):
    """ A SequenceConstraintFactory subclass that represents an abutment 
    constraint, which is a constraint between two anchors of different 
    components separated by some amount of space. It has a 'from_items'
    classmethod which will create a sequence of abutment constraints
    from a sequence of items, a direction, and default spacing.

    """
    #: A mapping from orientation to the order of anchor names to 
    #: lookup for a pair of items in order to make the constraint.
    orientation_map = {
        'horizontal': ('right', 'left'),
        'vertical': ('bottom', 'top'),
    }

    @classmethod
    def from_items(cls, items, orientation, spacing):
        """ A classmethod that generates a list of abutment constraints
        given a sequence of items, an orientation, and default spacing. 

        Parameters
        ----------
        items : sequence
            A valid sequence of constrainable objects. These inclue
            instances of Constrainable, LinearSymbolic, Spacer,
            and int.
        
        orientation : string
            Either 'vertical' or 'horizontal', which represents the 
            orientation in which to abut the items.
        
        spacing : int
            The spacing to use between items if no spacing is explicitly 
            provided by in the sequence of items.
        
        Returns
        -------
        result : list
            A list of AbutmentConstraint instances.
        
        Notes
        ------ 
        The order of abutment is left-to-right for horizontal direction 
        and top-to-bottom for vertical direction.

        """
        # Grab the tuple of anchor names to lookup for each pair of 
        # items in order to make the connection.
        orient = cls.orientation_map.get(orientation)
        if orient is None:
            msg = ("Valid orientations for abutment are 'vertical' or "
                   "'horizontal'. Got %r instead.")
            raise ValueError(msg % orientation)
        first_name, second_name = orient
        return cls._make_cns(items, first_name, second_name, spacing)


class AlignmentConstraintFactory(SequenceConstraintFactory):
    """ A SequenceConstraintFactory subclass which represents an 
    alignmnent constraint, which is a constraint between two anchors of
    different components which are aligned but may be separated by some
    amount of space. It provides a 'from_items' classmethod which will 
    create a  list of alignment constraints from a sequence of items an 
    anchor name, and a default spacing.

    """
    @classmethod
    def from_items(cls, items, anchor_name, spacing):
        """ A classmethod that will create a seqence of alignment
        constraints given a sequence of items, an anchor name, and
        a default spacing.
        
        Parameters
        ----------
        items : sequence
            A valid sequence of constrainable objects. These inclue
            instances of Constrainable, LinearSymbolic, Spacer,
            and int.
        
        anchor_name : string
            The name of the anchor on the components which should be
            aligned. Either 'left', 'right', 'top', 'bottom', 'v_center',
            or 'h_center'.
        
        spacing : int
            The spacing to use between items if no spacing is explicitly 
            provided by in the sequence of items.
        
        Returns
        -------
        result : list
            A list of AbutmentConstraint instances.

        Notes
        -----
        For every item in the sequence, if the item is a component, then
        anchor for the given anchor_name on that component will be used. 
        If a LinearSymbolic is given, then that symbolic will be used and
        the anchor_name will be ignored. Specifying space between items 
        via integers or spacers is allowed.

        """
        return cls._make_cns(items, anchor_name, anchor_name, spacing)


#------------------------------------------------------------------------------
# Spacers
#------------------------------------------------------------------------------
class Spacer(object):
    """ An abstract base class for spacers. Subclasses must implement 
    the 'constrain' method.

    """
    __metaclass__ = ABCMeta

    def __init__(self, amt, strength=None, weight=None):
        self.amt = max(0, amt)
        self.strength = strength
        self.weight = weight

    def when(self, switch):
        """ A simple method that can be used to switch off the generated
        space depending on a boolean value.

        """
        if switch:
            return self
    
    def constrain(self, first_anchor, second_anchor):
        """ Returns the list of generated constraints appropriately
        weighted by the default strength and weight, if provided.

        """
        constraints = self._constrain(first_anchor, second_anchor)
        strength = self.strength
        if strength is not None:
            constraints = [cn | strength for cn in constraints]
        weight = self.weight
        if weight is not None:
            constraints = [cn | weight for cn in constraints]
        return constraints

    @abstractmethod
    def _constrain(self, first_anchor, second_anchor):
        """ An abstract method. Subclasses should implement this method 
        to return a list of LinearConstraint instances which separate
        the two anchors according to the amount of space represented
        by the spacer.

        """
        raise NotImplementedError


class EqSpacer(Spacer):
    """ A spacer which represents a fixed amount of space.

    """
    def _constrain(self, first_anchor, second_anchor):
        """ A constraint of the form (anchor_1 + space == anchor_2)

        """
        return [(first_anchor + self.amt) == second_anchor]


class LeSpacer(Spacer):
    """ A spacer which represents a flexible space with a maximum value.

    """
    def _constrain(self, first_anchor, second_anchor):
        """ A constraint of the form (anchor_1 + space >= anchor_2)
        That is, the visible space must be less than or equal to the
        given amount. An additional constraint is applied which 
        constrains (anchor_1 <= anchor_2) to prevent negative space.

        """
        return [(first_anchor + self.amt) >= second_anchor,
                first_anchor <= second_anchor]


class GeSpacer(Spacer):
    """ A spacer which represents a flexible space with a minimum value.

    """
    def _constrain(self, first_anchor, second_anchor):
        """ A constraint of the form (anchor_1 + space <= anchor_2)
        That is, the visible space must be greater than or equal to
        the given amount.

        """
        return [(first_anchor + self.amt) <= second_anchor]


class FlexSpacer(Spacer):
    """ A spacer which represents a space with a hard minimum, but also 
    a weaker preference for being that minimum.

    """
    def __init__(self, amt, min_strength='required', min_weight=1.0, eq_strength='medium', eq_weight=1.25):
        self.amt = max(0, amt)
        self.min_strength = min_strength
        self.min_weight = min_weight
        self.eq_strength = eq_strength
        self.eq_weight = eq_weight

    def constrain(self, first_anchor, second_anchor):
        """ Return list of LinearConstraint objects that are appropriate to
        separate the two anchors according to the amount of space represented by
        the spacer.

        """
        return self._constrain(first_anchor, second_anchor)

    def _constrain(self, first_anchor, second_anchor):
        """ Constraints of the form (anchor_1 + space <= anchor_2) and
        (anchor_1 + space == anchor_2)

        """
        return [
            ((first_anchor + self.amt) <= second_anchor) | self.min_strength | self.min_weight,
            ((first_anchor + self.amt) == second_anchor) | self.eq_strength | self.eq_weight,
        ]


class LayoutSpacer(Spacer):
    """ A Spacer instance which supplies convenience symbolic and normal
    methods to facilitate specifying spacers in layouts.

    """
    def __call__(self, *args, **kwargs):
        return self.__class__(*args, **kwargs)
    
    def __eq__(self, other):
        if not isinstance(other, int):
            raise TypeError('space can only be created from ints')
        return EqSpacer(other, self.strength, self.weight)
    
    def __le__(self, other):
        if not isinstance(other, int):
            raise TypeError('space can only be created from ints')
        return LeSpacer(other, self.strength, self.weight)
    
    def __ge__(self, other):
        if not isinstance(other, int):
            raise TypeError('space can only be created from ints')
        return GeSpacer(other, self.strength, self.weight)
    
    def _constrain(self, first_anchor, second_anchor):
        """ Returns a greater than or equal to spacing constraint.

        """
        spacer = GeSpacer(self.amt, self.strength, self.weight)
        return spacer._constrain(first_anchor, second_anchor)
        
    def flex(self, **kwargs):
        """ Returns a flex spacer for the current amount.

        """
        return FlexSpacer(self.amt, **kwargs)


#------------------------------------------------------------------------------
# Layout factory functions
#------------------------------------------------------------------------------
def horizontal(*items, **metadata):
    """ Create a DeferredConstraints object composed of horizontal 
    abutments for the given sequence of items.

    """
    metadata['orientation'] = 'horizontal'
    return AbutmentHelper(*items, **metadata)


def vertical(*items, **metadata):
    """ Create a DeferredConstraints object composed of vertical 
    abutments for the given sequence of items.

    """
    metadata['orientation'] = 'vertical'
    return AbutmentHelper(*items, **metadata)


def hbox(*items, **metadata):
    """ Create a DeferredConstraints object composed of horizontal 
    abutments for a given sequence of items.

    """
    metadata['orientation'] = 'horizontal'
    return LinearBoxHelper(*items, **metadata)


def vbox(*items, **metadata):
    """ Create a DeferredConstraints object composed of vertical abutments 
    for a given sequence of items.

    """
    metadata['orientation'] = 'vertical'
    return LinearBoxHelper(*items, **metadata)


def align(anchor, *items, **metadata):
    """ Align the given anchors of the given components. Inter-component 
    spacing is allowed.

    """
    metadata.setdefault('anchor', anchor)
    return AlignmentHelper(*items, **metadata)


@deprecated
def align_left(*items, **metadata):
    """ Align the left anchors of the given components. Inter-component
    spacing is allowed.

    """
    return align('left', *items, **metadata)


@deprecated
def align_right(*items, **metadata):
    """ Align the right anchors of the given components. Inter-component
    spacing is allowed.
    
    """
    return align('right', *items, **metadata)


@deprecated
def align_top(*items, **metadata):
    """ Align the top anchors of the given components. Inter-component
    spacing is allowed.
    
    """
    return align('top', *items, **metadata)


@deprecated
def align_bottom(*items, **metadata):
    """ Align the bottom anchors of the given components. Inter-component
    spacing is allowed.
    
    """
    return align('bottom', *items, **metadata)


@deprecated
def align_v_center(*items, **metadata):
    """ Align the v_center anchors of the given components. Inter-component
    spacing is allowed.
    
    """
    return align('v_center', *items, **metadata)


@deprecated
def align_h_center(*items, **metadata):
    """ Align the h_center anchors of the given components. Inter-component
    spacing is allowed.
    
    """
    return align('h_center', *items, **metadata)


#------------------------------------------------------------------------------
# Toolkit items
#------------------------------------------------------------------------------
LAYOUT_HELPERS = {
    'horizontal': horizontal,
    'vertical': vertical,
    'hbox': hbox,
    'vbox': vbox,
    'align_left': align_left,
    'align_right': align_right,
    'align_top': align_top,
    'align_bottom': align_bottom,
    'align_v_center': align_v_center,
    'align_h_center': align_h_center,
    'align': align,
    '_space_': LayoutSpacer(DefaultSpacing.ABUTMENT), # Backwards compatibility
    'spacer': LayoutSpacer(DefaultSpacing.ABUTMENT),
}




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .wx_test_assistant import WXTestAssistant, skip_nonwindows
from .. import label


@skip_nonwindows
class TestWxLabel(WXTestAssistant, label.TestLabel):
    """ WXLabel tests. """

    def get_text(self, widget):
        """ Get a label's text.

        """
        return widget.GetLabel()



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import itertools
import operator

from .qt.QtCore import Qt, QAbstractItemModel, QModelIndex, QSize
from .styling import q_color_from_color, q_font_from_font

from ...item_models.abstract_item_model import AbstractItemModel


#: An invalid QModelIndex() for use in conversion routines
_INVALID_QINDEX = QModelIndex()


#------------------------------------------------------------------------------
# Flag Map
#------------------------------------------------------------------------------
# The construction cost of creating an ItemsFlag object is very high
# so instead we enumerate all 128 combinations so we can look them 
# up as a simple map. This is a hack to workaround a PySide issue
# where it doesn't accept ints as return values from the flags() method
_qitem_flags = (
    Qt.NoItemFlags, Qt.ItemIsSelectable, Qt.ItemIsEditable,
    Qt.ItemIsDragEnabled, Qt.ItemIsDropEnabled, Qt.ItemIsUserCheckable,
    Qt.ItemIsEnabled, Qt.ItemIsTristate,
)


_QITEM_FLAGS = {}


for n in range(1, len(_qitem_flags) + 1):
    for items in itertools.combinations(_qitem_flags, n):
        val = reduce(operator.or_, items, 0)
        _QITEM_FLAGS[int(val)] = val


del _qitem_flags


#------------------------------------------------------------------------------
# Role Mapping Builders
#------------------------------------------------------------------------------
def _build_getters(model):
    """ Returns a dictionary of role->getter methods for the given
    model.

    """
    return {
        int(Qt.DisplayRole): model.data,
        int(Qt.DecorationRole): model.decoration,
        int(Qt.EditRole): model.edit_data,
        int(Qt.ToolTipRole): model.tool_tip,
        int(Qt.StatusTipRole): model.status_tip,
        int(Qt.WhatsThisRole): model.whats_this,
        int(Qt.FontRole): model.font,
        int(Qt.TextAlignmentRole): model.alignment,
        int(Qt.BackgroundRole): model.background,
        int(Qt.ForegroundRole): model.foreground,
        int(Qt.CheckStateRole): model.check_state,
        int(Qt.SizeHintRole): model.size_hint,
        int(Qt.UserRole): lambda *args, **kwargs: None,
    }


def _build_setters(model):
    """ Returns a dictionary of role->setter methods for the given
    model.

    """
    return {
        int(Qt.DisplayRole): model.set_data,
        int(Qt.EditRole): model.set_data,
        int(Qt.CheckStateRole): model.set_check_state,
    }


def _build_h_header_getters(model):
    """ Returns a dictionary of role->h_header getter methods for the
    given model.

    """
    return {
        int(Qt.DisplayRole): model.horizontal_header_data,
        int(Qt.DecorationRole): model.horizontal_header_decoration,
        int(Qt.ToolTipRole): model.horizontal_header_tool_tip,
        int(Qt.StatusTipRole): model.horizontal_header_status_tip,
        int(Qt.WhatsThisRole): model.horizontal_header_whats_this,
        int(Qt.FontRole): model.horizontal_header_font,
        int(Qt.TextAlignmentRole): model.horizontal_header_alignment,
        int(Qt.BackgroundRole): model.horizontal_header_background,
        int(Qt.ForegroundRole): model.horizontal_header_foreground,
        int(Qt.SizeHintRole): model.horizontal_header_size_hint,
        int(Qt.UserRole): lambda *args, **kwargs: None,
    }


def _build_v_header_getters(model):
    """ Returns a dictionary of role->v_header getter methods for the
    given model.

    """
    return {
        int(Qt.DisplayRole): model.vertical_header_data,
        int(Qt.DecorationRole): model.vertical_header_decoration,
        int(Qt.ToolTipRole): model.vertical_header_tool_tip,
        int(Qt.StatusTipRole): model.vertical_header_status_tip,
        int(Qt.WhatsThisRole): model.vertical_header_whats_this,
        int(Qt.FontRole): model.vertical_header_font,
        int(Qt.TextAlignmentRole): model.vertical_header_alignment,
        int(Qt.BackgroundRole): model.vertical_header_background,
        int(Qt.ForegroundRole): model.vertical_header_foreground,
        int(Qt.SizeHintRole): model.vertical_header_size_hint,
        int(Qt.UserRole): lambda *args, **kwargs: None,
    }


#------------------------------------------------------------------------------
# Data Converters
#------------------------------------------------------------------------------
def _not_yet_supported(val):
    # use this converter for roles that are not 
    # yet supported.
    return None


def _no_convert(val):
    # These values don't need converting because they should
    # come from the user model with a Qt compatible type.
    return val


def _brush_convert(brush):
    # We don't yet support a full brush, so for now just 
    # return the color. Which works fine.
    if brush:
        return q_color_from_color(brush.color)


def _size_convert(size):
    if size is not None:
        return QSize(*size)

# Some of these should be LRU cached    
def _font_convert(font):
    if font:
        return q_font_from_font(font)


_QROLE_CONVERTERS = {
    int(Qt.DisplayRole): _no_convert,
    int(Qt.DecorationRole): _not_yet_supported,
    int(Qt.EditRole): _no_convert,
    int(Qt.ToolTipRole): _no_convert,
    int(Qt.StatusTipRole): _no_convert,
    int(Qt.WhatsThisRole): _no_convert,
    int(Qt.FontRole): _font_convert,
    int(Qt.TextAlignmentRole): _no_convert,
    int(Qt.BackgroundRole): _brush_convert,
    int(Qt.ForegroundRole): _brush_convert,
    int(Qt.CheckStateRole): _not_yet_supported,
    int(Qt.SizeHintRole): _size_convert,
    int(Qt.UserRole): _not_yet_supported,
}


#------------------------------------------------------------------------------
# AbstractItemModelWrapper
#------------------------------------------------------------------------------
class AbstractItemModelWrapper(QAbstractItemModel):
    """ A QAbstractItemModel subclass which wraps an instance of an
    Enaml AbstractItemModel and does the appropriate delegation and
    conversion between Python and Qt

    """
    def __init__(self, item_model):
        super(AbstractItemModelWrapper, self).__init__()
        if not isinstance(item_model, AbstractItemModel):
            raise TypeError('Model must be an instance of AbstractItemModel.')
        
        self._item_model = item_model
        self._getters = _build_getters(item_model)
        self._setters = _build_setters(item_model)
        self._h_header_getters = _build_h_header_getters(item_model)
        self._v_header_getters = _build_v_header_getters(item_model)

        def listen(name):
            signal = getattr(item_model, name)
            signal.connect(getattr(self, '_' + name))

        listen('columns_about_to_be_inserted')
        listen('columns_about_to_be_moved')
        listen('columns_about_to_be_removed')
        listen('columns_inserted')
        listen('columns_moved')
        listen('columns_removed')

        listen('rows_about_to_be_inserted')
        listen('rows_about_to_be_moved')
        listen('rows_about_to_be_removed')
        listen('rows_inserted')
        listen('rows_moved')
        listen('rows_removed')

        listen('layout_about_to_be_changed')
        listen('layout_changed')

        listen('model_about_to_be_reset')
        listen('model_reset')
        listen('data_changed')
        listen('horizontal_header_data_changed')
        listen('vertical_header_data_changed')

    #--------------------------------------------------------------------------
    # Traits Event Handlers
    #--------------------------------------------------------------------------
    def _columns_about_to_be_inserted(self, evt_arg):
        parent, start, end = evt_arg
        q_index = self.to_q_index(parent)
        self.columnsAboutToBeInserted.emit(q_index, start, end)
    
    def _columns_about_to_be_moved(self, evt_arg):
        parent, start, end = evt_arg
        q_index = self.to_q_index(parent)
        self.columnsAboutToBeMoved.emit(q_index, start, end)
    
    def _columns_about_to_be_removed(self, evt_arg):
        parent, start, end = evt_arg
        q_index = self.to_q_index(parent)
        self.columnsAboutToBeRemoved.emit(q_index, start, end)
    
    def _columns_inserted(self, evt_arg):
        parent, start, end = evt_arg
        q_index = self.to_q_index(parent)
        self.columnsInserted.emit(q_index, start, end)
    
    def _columns_moved(self, evt_arg):
        parent, start, end = evt_arg
        q_index = self.to_q_index(parent)
        self.columnsMoved.emit(q_index, start, end)
    
    def _columns_removed(self, evt_arg):
        parent, start, end = evt_arg
        q_index = self.to_q_index(parent)
        self.columnsRemoved.emit(q_index, start, end)

    def _rows_about_to_be_inserted(self, evt_arg):
        parent, start, end = evt_arg
        q_index = self.to_q_index(parent)
        self.rowsAboutToBeInserted.emit(q_index, start, end)
    
    def _rows_about_to_be_moved(self, evt_arg):
        parent, start, end = evt_arg
        q_index = self.to_q_index(parent)
        self.rowsAboutToBeMoved.emit(q_index, start, end)
    
    def _rows_about_to_be_removed(self, evt_arg):
        parent, start, end = evt_arg
        q_index = self.to_q_index(parent)
        self.rowsAboutToBeRemoved.emit(q_index, start, end)
    
    def _rows_inserted(self, evt_arg):
        parent, start, end = evt_arg
        q_index = self.to_q_index(parent)
        self.rowsInserted.emit(q_index, start, end)
    
    def _rows_moved(self, evt_arg):
        parent, start, end = evt_arg
        q_index = self.to_q_index(parent)
        self.rowsMoved.emit(q_index, start, end)
    
    def _rows_removed(self, evt_arg):
        parent, start, end = evt_arg
        q_index = self.to_q_index(parent)
        self.rowsRemoved.emit(q_index, start, end)
    
    def _layout_about_to_be_changed(self):
        self.layoutAboutToBeChanged.emit()
    
    def _layout_changed(self):
        self.layoutChanged.emit()
            
    def _model_about_to_be_reset(self):
        self.modelAboutToBeReset.emit()
    
    def _model_reset(self):
        self.modelReset.emit()
    
    def _data_changed(self, evt_arg):
        top_left, bottom_right = evt_arg
        q_top_left = self.to_q_index(top_left)
        q_bottom_right = self.to_q_index(bottom_right)
        self.dataChanged.emit(q_top_left, q_bottom_right)
    
    def _horizontal_header_data_changed(self, evt_arg):
        first, last = evt_arg
        self.headerDataChanged.emit(Qt.Horizontal, first, last)

    def _vertical_header_data_changed(self, evt_arg):
        first, last = evt_arg
        self.headerDataChanged.emit(Qt.Vertical, first, last)
    
    #--------------------------------------------------------------------------
    # AbstractItemModel interface 
    #--------------------------------------------------------------------------
    def buddy(self, index):
        enaml_index = self.from_q_index(index)
        enaml_buddy = self._item_model.buddy(enaml_index)
        return self.to_q_index(enaml_buddy)

    def canFetchMore(self, parent):
        enaml_parent = self.from_q_index(parent)
        return self._item_model.can_fetch_more(enaml_parent)

    def fetchMore(self, parent):
        enaml_parent = self.from_q_index(parent)
        return self._item_model.fetchMore(enaml_parent)
        
    def headerData(self, section, orientation, role):
        if orientation == Qt.Horizontal:
            res = self._h_header_getters[role](section)
        else:
            res = self._v_header_getters[role](section)
        return res

    def setData(self, index, value, role):
        enaml_index = self.from_q_index(index)
        setter = self._setters.get(role)
        if setter is not None:
            res = setter(enaml_index, value)
        else:
            res = False
        return res
        
    def flags(self, index):
        enaml_index = self.from_q_index(index)
        enaml_flags = self._item_model.flags(enaml_index)
        return _QITEM_FLAGS[enaml_flags]

    def rowCount(self, parent):
        enaml_index = self.from_q_index(parent)
        return self._item_model.row_count(enaml_index)

    def columnCount(self, parent):
        enaml_index = self.from_q_index(parent)
        return self._item_model.column_count(enaml_index)
    
    def index(self, row, column, parent):
        enaml_parent = self.from_q_index(parent)
        enaml_index = self._item_model.index(row, column, enaml_parent)
        return self.to_q_index(enaml_index)
    
    def parent(self, index):
        enaml_index = self.from_q_index(index)
        enaml_parent = self._item_model.parent(enaml_index)
        return self.to_q_index(enaml_parent)

    def data(self, index, role):
        enaml_index = self.from_q_index(index)
        if enaml_index is None:
            return
        data = self._getters[role](enaml_index)
        return _QROLE_CONVERTERS[role](data)

    # XXX we don't have a use case at the moment for setHeaderData
    def setHeaderData(self, section, orientation, value, role):
       return False
    
    def from_q_index(self, q_index):
        if not q_index.isValid():
            return None
        row = q_index.row()
        col = q_index.column()
        context = q_index.internalPointer()
        return self._item_model.create_index(row, col, context)

    def to_q_index(self, enaml_index):
        if enaml_index is None:
            return _INVALID_QINDEX
        row = enaml_index.row
        col = enaml_index.column
        context = enaml_index.context
        return self.createIndex(row, col, context)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import wx


def wx_color_from_color(color):
    """ Converts an Enaml Color into a wxColor.

    """
    if not color:
        res = wx.NullColor
    else:
        res = wx.Color(*color)
    return res


font_style_map = {
    'normal': wx.FONTSTYLE_NORMAL,
    'italic': wx.FONTSTYLE_ITALIC,
    'oblique': wx.FONTSTYLE_SLANT,
}


font_family_hint_map = {
    'any': wx.FONTFAMILY_DEFAULT,
    'sans_serif': wx.FONTFAMILY_SWISS,
    'helvetica': wx.FONTFAMILY_DEFAULT,
    'times': wx.FONTFAMILY_ROMAN,
    'type_writer': wx.FONTFAMILY_TELETYPE,
    'courier': wx.FONTFAMILY_MODERN,
    'old_english': wx.FONTFAMILY_DEFAULT,
    'decorative': wx.FONTFAMILY_DECORATIVE,
    'monospace': wx.FONTFAMILY_MODERN,
    'fantasy': wx.FONTFAMILY_DEFAULT,
    'cursive': wx.FONTFAMILY_SCRIPT,
    'system': wx.FONTFAMILY_DEFAULT
}


# A cache of the default font which is created on demand
_cached_wx_default_font = None
def _wx_default_font():
    global _cached_wx_default_font
    if _cached_wx_default_font is None:
        font = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        _cached_wx_default_font = font
    return _cached_wx_default_font


def wx_font_from_font(font):
    """ Converts an Enaml Font into a wxFont.

    """
    point_size = font.point_size
    if point_size != -1:
        wx_point_size = point_size
    else:
        wx_point_size = _wx_default_font().GetPointSize()

    wx_family = font_family_hint_map[font.family_hint]
    
    wx_style = font_style_map[font.style]

    weight = font.weight
    if weight < 45:
        wx_weight = wx.FONTWEIGHT_LIGHT
    elif weight > 55:
        wx_weight = wx.FONTWEIGHT_BOLD
    else:
        wx_weight = wx.FONTWEIGHT_NORMAL

    wx_underline = font.underline

    wx_face = font.family
    if not wx_face:
        wx_face = _wx_default_font().GetFaceName()

    wx_font = wx.Font(
        wx_point_size,
        wx_family,
        wx_style,
        wx_weight,
        wx_underline,
        wx_face
    )

    return wx_font




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .enaml_test_case import EnamlTestCase, required_method


class TestPushButton(EnamlTestCase):
    """ Logic for testing push buttons.

    Tooklit testcases need to provide the following methods

    Abstract Methods
    ----------------
    button_pressed(self)
        Press the button programmatically.

    button_released(self)
        Release the button programmatically.

    button_clicked(self)
        Click the button programmatically.

    """

    def setUp(self):
        """ Set up before push button tests.

        """

        enaml_source = """
enamldef MainView(MainWindow):
    attr events
    PushButton:
        name = 'pb1'
        text = 'foo'
        clicked :: events.append('clicked')
        pressed :: events.append('pressed')
        released :: events.append('released')
"""

        self.events = []
        self.view = self.parse_and_create(enaml_source, events=self.events)
        self.component = self.component_by_name(self.view, 'pb1')
        self.widget = self.component.toolkit_widget

    def test_button_pressed(self):
        """ React to a push button press event.

        """
        self.button_pressed()

        events = self.events
        self.assertTrue(self.component.down)
        self.assertIn('pressed', events)
        self.assertNotIn('released', events)
        self.assertNotIn('clicked', events)

    def test_button_clicked(self):
        """ Test a push button click event.

        """
        self.button_clicked()

        events = self.events
        self.assertIn('clicked', events)
        self.assertNotIn('pressed', events)
        self.assertNotIn('released', events)

    def test_button_released(self):
        """ Test a push button release event.

        """
        events = self.events
        # Release events are ignored if the button weas not already down
        self.button_released()
        self.assertEqual(events, [])

    def test_press_release_sequence(self):
        """ Verify the even firing when the press-release (nornal) sequence
        is applied.

        """
        events = self.events

        self.button_pressed()
        self.assertTrue(self.component.down)
        self.button_released()
        self.assertFalse(self.component.down)
        self.assertEqual(events.count('pressed'), 1)
        self.assertEqual(events.count('released'), 1)
        self.assertNotIn('clicked', events)

    def test_button_all_events(self):
        """ Test press, release, and click events.

        """
        self.button_pressed()
        self.button_released()
        self.button_clicked()

        events = self.events
        self.assertEqual(events.count('clicked'), 1)
        self.assertEqual(events.count('pressed'), 1)
        self.assertEqual(events.count('released'), 1)

    def test_button_down(self):
        """ Test the button's `down` attribute.

        """
        component = self.component

        self.assertFalse(component.down)
        self.button_pressed()
        self.assertTrue(component.down)
        self.button_released()
        self.assertFalse(component.down)

    #--------------------------------------------------------------------------
    # Abstract methods
    #--------------------------------------------------------------------------
    @required_method
    def button_pressed(self):
        """ Press the button programmatically.

        """
        pass

    @required_method
    def button_released(self):
        """ Release the button programmatically.

        """
        pass

    @required_method
    def button_clicked(self):
        """ Click the button programmatically.

        """
        pass




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import abstractmethod

from traits.api import Property, List, Instance, cached_property

from .layout_task_handler import LayoutTaskHandler
from .menu import Menu
from .widget_component import WidgetComponent, AbstractTkWidgetComponent


class AbstractTkMenuBar(AbstractTkWidgetComponent):
    """ The abstract toolkit interface for a MenuBar.

    """
    @abstractmethod
    def shell_menus_changed(self, menus):
        """ Update the menu bar with the new list of menu objects.

        """
        raise NotImplementedError


class MenuBar(LayoutTaskHandler, WidgetComponent):
    """ A declarative Enaml Component which represents a menu bar in
    a main window.

    """
    #: A read-only cached property which holds the list of children
    #: which are instances of Menu.
    menus = Property(List(Instance(Menu)), depends_on='children')

    #: Overridden parent class trait
    abstract_obj = Instance(AbstractTkMenuBar)

    @cached_property
    def _get_menus(self):
        """ The property getter for the 'menus' attribute.

        """
        flt = lambda child: isinstance(child, Menu)
        return filter(flt, self.children)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt_test_assistant import QtTestAssistant
from .. import progress_bar


class TestQtProgressBar(QtTestAssistant, progress_bar.TestProgressBar):
    """ QtProgressBar tests.
    
    """

    def get_value(self, widget):
        """  Get the toolkits widget's active value.

        """
        return widget.value()

    def get_minimum(self, widget):
        """  Get the toolkits widget's maximum value attribute.

        """
        return widget.minimum()

    def get_maximum(self, widget):
        """ Get the toolkits widget's minimum value attribute.

        """
        return widget.maximum()



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .abstract_locale import AbstractLocale
from .system_locale import SystemLocale


#: The storage for the framework default locale object
_default_locale = None


def set_default_locale(locale_obj):
    """ Sets the framework-wide default locale.

    Parameters
    ----------
    locale_obj : AbstractLocale
        An instance of AbstractLocale which will be used as the default
        locale for the Enaml framework.

    """
    global _default_locale
    if locale_obj is not None and not isinstance(locale_obj, AbstractLocale):
        msg = ('Locale object must be an instance of AbstractLocale or None. '
               'Got %s instead.' % type(locale_obj))
        raise TypeError(msg)
    _default_locale = locale_obj


def get_default_locale():
    """ Returns the framework-wide default locale object. If a default
    has not been explicitly provided, this will return the default
    SystemLocale

    """
    locale_obj = _default_locale
    if locale_obj is None:
        locale_obj = SystemLocale.default()
        set_default_locale(locale_obj)
    return locale_obj




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from traits.api import List, Instance, Either, Bool, Property, cached_property

from ..core.base_component import BaseComponent


class Include(BaseComponent):
    """ A BaseComponent subclass which allows children to be dynamically
    included into a parent.

    """
    #: A read-only property which returns the toolkit widget for this
    #: component. This call is proxied to the parent and will return
    #: None if the parent does not have a toolkit widget defined. 
    #: It is necessary to proxy the toolkit_widget so that any child
    #: Include components can access the proper toolkit widget to 
    #: pass down to their dynamic components.
    toolkit_widget = Property

    #: The dynamic components of the Include. This is a cached property
    #: which will accept a single component, or a list of components as
    #: input. If the input is a single component, it will be converted
    #: into a single element list.
    components = Property(
        Either(List(Instance(BaseComponent)), Instance(BaseComponent)),
        depends_on='_components',
    )
    
    #: A private attribute which stores the underlying list of created
    #: components. This list should not be manipulated by user code.
    _components = List(Instance(BaseComponent))

    #: A private Boolean flag indicating if the dynamic components of 
    #: this Include have been intialized for the first time. This should 
    #: not be manipulated directly by the user.
    _components_initialized = Bool(False)

    #--------------------------------------------------------------------------
    # Property Getters and Setters
    #--------------------------------------------------------------------------
    @cached_property
    def _get_components(self):
        """ The cached property getter for the 'components' attribute. 
        It simply returns the private '_components' list.

        """
        return self._components
    
    def _set_components(self, val):
        """ The property setter for the 'components' attribute. It will
        convert a single component into a single element list.

        """
        if isinstance(val, BaseComponent):
            val = [val]
        self._components = val
    
    def _get_toolkit_widget(self):
        """ The property getter for the 'toolkit_widget' attribute.

        """
        try:
            res = self.parent.toolkit_widget
        except AttributeError:
            res = None
        return res

    #--------------------------------------------------------------------------
    # Setup Methods 
    #--------------------------------------------------------------------------
    def _setup_init_layout(self):
        """ A reimplemented parent class method which builds the initial
        list of include components during the layout initialization pass.
        The layout is performed bottom-up so that the tree is up-to-date
        before any parents compute their layout.

        """
        super(Include, self)._setup_init_layout()
        self._setup_components()
        self.on_trait_change(
            self._on_components_actual_updated, '_components:_actual_updated',
        )
        self._actual_updated()

    def _setup_components(self):
        """ An internal method used to setup the dynamic child components.

        """
        cmpnts = self.components
        if not cmpnts:
            return

        # The dynamic components of an Include are injected into the 
        # parent's children inline so that they *appear* as if they
        # are siblings of the Include. To do this, we pass the reference
        # to the parent of the Include and the corresponding toolkit
        # widget where appropriate.
        parent = self.parent
        try:
            # If our parent is a BaseComponent, it won't have a 
            # toolkit_widget attribute.
            toolkit_parent = parent.toolkit_widget
        except AttributeError:
            toolkit_parent = None

        # The following blocks perform roughly the same setup process
        # as BaseComponent.setup(), except that we don't need to perform
        # the setup for this Include instance (since it's already setup).
        # Also, since we don't need to recurse into any children (an
        # Include can't have children), there is no need to break these
        # blocks out into separate methods.

        # Need to explicitly assign the parent to the components 
        # since they were not added via the add_subcomponent method.
        for child in cmpnts:
            child.parent = self
        
        for child in cmpnts:
            child._setup_create_widgets(toolkit_parent)
            
        for child in cmpnts:
            child._setup_init_widgets()

        for child in cmpnts:
            child._setup_eval_expressions()

        for child in cmpnts:
            child._setup_bind_widgets()
                
        for child in cmpnts:
            child._setup_listeners()

        for child in cmpnts:
            child._setup_init_visibility()

        for child in cmpnts:
            child._setup_init_layout()

        for child in cmpnts:
            child._setup_finalize()

        for child in cmpnts:
            child._setup_set_initialized()

        self._components_initialized = True

    #--------------------------------------------------------------------------
    # Parent Class Overrides 
    #--------------------------------------------------------------------------
    def add_subcomponent(self, component):
        """ An overriden parent class method which prevents subcomponents
        from being declared for an Include instance.

        """
        msg = ("Cannot add subcomponents to an Include. Assign a component "
               "or a list components to the 'components' attribute instead.")
        raise ValueError(msg)

    def get_actual(self):
        """ A reimplemented parent class method to include the dynamic
        children of this Include in our parent's list of children.

        """
        if self._components_initialized:
            res = sum([c.get_actual() for c in self.components], [])
        else:
            res = []
        return res

    def destroy(self):
        """ A re-implemented parent class method which destroys all of
        the underlying dynamic children.

        """
        super(Include, self).destroy()
        for item in self.components:
            item.destroy()

    #--------------------------------------------------------------------------
    # Change Handlers 
    #--------------------------------------------------------------------------
    def _components_changed(self, name, old, new):
        """ Reacts to changes in the dynamic components and sets up the 
        new children, making sure the old ones are destroyed and that 
        the '_actual_updated' event gets fired.

        """
        # The first time a cached property is set, the notification
        # handler will be sent None as the old value even if the 
        # property getter points to something that has a value. In 
        # all other cases, it works as expected. We guard against
        # that condition as well a making sure the object is fully
        # initialized before destroying any of the old children.
        if self.initialized and old is not None:
            def closure():
                for item in old:
                    item.destroy()
                self._setup_components()
                self._actual_updated()
            self.request_relayout_task(closure)
            
    # This notifier is hooked up in the '_handle_initialized' method 
    # due to issues surrounding trait_setq contexts.
    def _on_components_actual_updated(self):
        """ Handles a '_actual_updated' event being fired by one 
        the dynamic components. The event is proxied up the tree by
        firing the same event on this instance. This allows a nested
        Include to update its contents independent of the Include in
        which it is nested.

        """        
        self._actual_updated()




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import abstractmethod

from traits.api import HasTraits, Instance
from traitsui.api import View, Handler

from .control import Control, AbstractTkControl


class AbstractTkTraitsUIItem(AbstractTkControl):
    
    @abstractmethod
    def shell_model_changed(self, model):
        raise NotImplementedError
    
    @abstractmethod
    def shell_view_changed(self, view):
        raise NotImplementedError
    
    @abstractmethod
    def shell_handler_changed(self, view):
        raise NotImplementedError


class TraitsUIItem(Control):
    """ A control which embeds a traits ui view.

    The TraitsUIItem allows a traits ui window to be embedded into an
    Enaml application. The ui that is embedded will be the same as that 
    which is created by calling model.edit_traits(...).

    """
    #: The model being displayed in the traits view.
    model = Instance(HasTraits)

    #: An optional specialized view to use for the model.
    view = Instance(View)

    #: The optional handler to use for the model and view.
    handler = Instance(Handler)

    #: Overridden parent class trait
    abstract_obj = Instance(AbstractTkTraitsUIItem)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from collections import deque
from threading import Lock

from .qt.QtCore import  Qt, QObject, QTimer, QThread, Slot, QMetaObject
from .qt.QtGui import QApplication

from ...components.abstract_application import AbstractTkApplication


class DeferredCaller(QObject):
    """ A QObject subclass that will invoke a callable on the main
    gui event thread. The only public api is the enqueue classmethod.

    """
    #: The private internal queue of pending tasks.
    _queue = deque()

    #: The private threading Lock which protects access to the queue 
    #: and creation of the internal singleton instance.
    _lock = Lock()

    @classmethod
    def enqueue(cls, callback, *args, **kwargs):
        """ Invoke the given callable in the main gui event thread at 
        some point in the future.

        Parameters
        ----------
        callback : callable
            The callable object to execute at some point in the future.

        *args
            Any positional arguments to pass to the callback.

        **kwargs
            Any keyword arguments to pass to the callback.

        """
        try:
            instance = cls._instance
        except AttributeError:
            with cls._lock:
                try:
                    instance = cls._instance
                except AttributeError:
                    instance = cls._instance = cls()
        instance._enqueue(callback, args, kwargs)

    def __init__(self):
        super(DeferredCaller, self).__init__()
        # Move the object to the main thread so that all signal/slot 
        # connections invoke the slots on the main thread. This call
        # is required since there is potential for this object to be
        # created from a non-gui thread.
        self.moveToThread(QApplication.instance().thread())

    def _enqueue(self, callback, args, kwargs):
        """ A private method which places the callback, args, and kwargs
        into the internal queue and and invokes the dispatch method on 
        the main thread.

        """
        item = (callback, args, kwargs)
        with self._lock:
            self._queue.append(item)
        # Invoking the method via QMetaObject with a QueuedConnection
        # has seemed to be the most reliable way of executing something
        # on the main thread. Other attempts with using QEvents under
        # heavy load have resulted in dropped (maybe compressed?)
        # events which leads to hard to trace bugs when application
        # code is relying on callbacks to be executed.
        QMetaObject.invokeMethod(self, '_dispatch', Qt.QueuedConnection)

    @Slot()
    def _dispatch(self):
        """ A private method which runs the next task in the queue. This
        method is a qt slot and is invoked from the main gui thread via
        QMetaObject.invokMethod.
        
        """ 
        queue = self._queue
        with self._lock:
            if not queue:
                return
            callback, args, kwargs = queue.popleft()
        callback(*args, **kwargs)


class QtApplication(AbstractTkApplication):
    """ A Qt4 implementation of AbstractTkApplication.

    """
    def initialize(self, *args, **kwargs):
        """ Initializes the underlying QApplication object. It does 
        *not* start the event loop. If the application object is already
        created, it is a no-op.

        """     
        app = QApplication.instance()
        if app is None:
            if not args:
                args = ([''],)
            app = QApplication(*args, **kwargs)   
        self._qt_app = app
    
    def start_event_loop(self):
        """ Starts the underlying application object's event loop, or 
        does nothing if it is already started. A RuntimeError will be 
        raised if the application object is not yet created.

        """
        app = QApplication.instance()
        if app is None:
            msg = 'Cannot start event loop. Application object not created.'
            raise RuntimeError(msg)
        
        if not getattr(app, '_in_event_loop', False):
            app._in_event_loop = True
            app.exec_()
            app._in_event_loop = False
    
    def event_loop_running(self):
        """ Returns True if the main event loop is running, False 
        otherwise.

        """
        app = QApplication.instance()
        return getattr(app, '_in_event_loop', False)

    def app_object(self):
        """ Returns the underlying application object, or None if one 
        does not exist.

        """
        return QApplication.instance()

    def is_main_thread(self):
        """ Return True if this method was called from the main event
        thread, False otherwise.

        """
        app = QApplication.instance()
        if app is None:
            raise RuntimeError('Application object not yet created')
        return app.thread() == QThread.currentThread()

    def call_on_main(self, callback, *args, **kwargs):
        """ Invoke the given callable in the main gui event thread at 
        some point in the future.

        Parameters
        ----------
        callback : callable
            The callable object to execute at some point in the future.

        *args
            Any positional arguments to pass to the callback.

        **kwargs
            Any keyword arguments to pass to the callback.

        """
        DeferredCaller.enqueue(callback, *args, **kwargs)

    def timer(self, ms, callback, *args, **kwargs):
        """ Invoke the given callable in the main gui event thread at 
        the given number of milliseconds in the future.

        Parameters
        ----------
        ms : int
            The number of milliseconds in the future to invoke the
            callable.

        callback : callable
            The callable object to execute at some point in the future.

        *args
            Any positional arguments to pass to the callback.

        **kwargs
            Any keyword arguments to pass to the callback.

        """
        fn = lambda: callback(*args, **kwargs)
        DeferredCaller.enqueue(QTimer.singleShot, ms, fn)

    def process_events(self):
        """ Process all of the pending events in the event queue.

        """
        QApplication.instance().processEvents()




#------------------------------------------------------------------------------
#  Copyright (c) 2012, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod, abstractproperty


class AbstractTkDockManager(object):
    """ 

    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, *dock_panes):
        """ Initialize a DockManager instance.

        Parameters
        ----------
        *dock_panes
            The initial DockPane instances to use with the manager.

        """
        raise NotImplementedError

    @abstractproperty
    def main_window(self):
        """ A property which gets and sets the manager's internal 
        reference to the MainWindow instance. The internal reference
        should be stored weakly.

        """ 
        raise NotImplementedError

    @abstractmethod
    def add_pane(self, pane):
        """ Add a dock pane to the manager and hence the window.

        Parameters
        ----------
        pane : DockPane
            The DockPane instance to add to the manager. If the pane is
            already being managed, this method call is a no-op.
        
        """
        raise NotImplementedError
    
    @abstractmethod
    def remove_pane(self, pane):
        """ Remove a dock pane from the manager and hence the window.

        Parameters
        ----------
        pane : DockPane
            The DockPane instance to remove from the manager. If the
            pane is not being managed, this method call is a no-op.
        
        """
        raise NotImplementedError

    @abstractmethod
    def panes(self):
        """ Get the list of dock panes currently being managed.

        Returns
        -------
        result : list
            The list of DockPane instance being managed by this manager.
        
        """
        raise NotImplementedError




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt_control import QtControl

from ...components.bounded_datetime import AbstractTkBoundedDatetime


class QtBoundedDatetime(QtControl, AbstractTkBoundedDatetime):
    """ A base class for use with widgets implementing behavior
    for subclasses of BoundedDatetime.

    """
    #--------------------------------------------------------------------------
    # Setup methods
    #--------------------------------------------------------------------------
    def initialize(self):
        super(QtBoundedDatetime, self).initialize()
        shell = self.shell_obj
        self.set_min_datetime(shell.min_datetime)
        self.set_max_datetime(shell.max_datetime)
        self.set_datetime(shell.datetime)

    #--------------------------------------------------------------------------
    # Abstract implementation methods
    #--------------------------------------------------------------------------
    def shell_datetime_changed(self, datetime):
        self.set_datetime(datetime)
    
    def shell_min_datetime_changed(self, min_datetime):
        self.set_min_datetime(min_datetime)
    
    def shell_max_datetime_changed(self, max_datetime):
        self.set_max_datetime(max_datetime)

    #--------------------------------------------------------------------------
    # Widget modification methods
    #--------------------------------------------------------------------------
    def set_datetime(self, datetime):
        raise NotImplementedError
    
    def set_min_datetime(self, min_datetime):
        raise NotImplementedError
    
    def set_max_datetime(self, max_datetime):
        raise NotImplementedError

